/////////////////////////////////////////////////////////////////////////
//	"devparam.cpp"
//------------------------------------------------------------------------
//	$Id: devparam.cpp 145 2005-12-19 06:03:27Z taro $
//////////////////////////////////////////////////////////////////////////

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>
#include	<time.h>
#include	"devtest.h"
#include	"defs.h"

#if defined(_DEBUG) && defined(WIN32)
# include	<crtdbg.h>
# define	DBGREPORT(param_)	_CrtDbgReport param_
#else
# define	DBGREPORT(param_)
#endif

DevTest::DevParam::DevParam()
{
	this->devname        = NULL;
	this->desc           = NULL;
	this->scsi           = FALSE;
	this->inqData        = NULL;
	this->capa           = NULL;
	this->pathId         = 0;
#if		defined(WIN32)
	this->targetId       = -1;	// scan
	this->lun            = -1;	// scan
#elif	defined(LINUX)
	this->targetId       = 0;
	this->lun            = 0;
#endif
	this->setDiskinf     = FALSE;

	this->file.path      = NULL;
#if		defined(WIN32)
	this->file.handle    = INVALID_HANDLE_VALUE;
#elif	defined(LINUX)
	this->file.fd        = -1;
#endif
	this->file.blockSize = 0;
	this->file.blockNum  = 0;
}

DevTest::DevParam::~DevParam()
{
	if ( this->devname != NULL )
		delete [] this->devname;
	if ( this->desc != NULL )
		delete [] this->desc;

	if ( this->capa != NULL )
		delete [] this->capa;

#if		defined(WIN32)
	if ( this->file.handle  != INVALID_HANDLE_VALUE )
		::CloseHandle(this->file.handle);
#elif	defined(LINUX)
	if ( this->file.fd != -1 )
		::close(this->file.fd);
#endif
	if ( this->file.path != NULL )
		delete [] this->file.path;
}

int		DevTest::DevParam::SetDevName(const char * text)
{
	if ( this->devname != NULL )
	{
		delete [] this->devname;
		this->devname = NULL;
	}
	if ( text != NULL )
	{
		if ( (this->devname = new char [strlen(text)+1]) == NULL )
			return -1;
		strcpy(this->devname, text);
	}
	return 0;
}

int		DevTest::DevParam::SetDesc(const char * text)
{
	if ( this->desc != NULL )
	{
		delete [] this->desc;
			this->desc = NULL;
	}
	if ( text != NULL )
	{
		if ( (this->desc = new char [strlen(text)+1]) == NULL )
			return -1;
		strcpy(this->desc, text);
	}
	return 0;
}

const char *	DevTest::DevParam::GetDesc(unsigned max)
{
	unsigned	len;

	if ( this->desc == NULL )
		return "";
	if ( max == 0 || (len = strlen(this->desc)) <= max )
		return this->desc;
	return &this->desc[len-max];
}

int	DevTest::DevParam::SetCapacityData(DevCtl::SCSICAPACITYDATA & buf)
{
	if ( this->capa == NULL )
	{
		if ( (this->capa = new DevCtl::SCSICAPACITYDATA) == NULL )
			return -1;
	}
	*(this->capa) = buf;
	return 0;
}

void	DevTest::DevParam::ClearCapacityData(void)
{
	if ( this->capa != NULL )
	{
		delete this->capa;
		this->capa = NULL;
	}
}

int		DevTest::DevParam::FileOpen(const char * path, INT64 blockNum, BOOL async)
{
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "DevParam::FileOpen path=%s\n", path));
	if ( this->CheckCapacityData() == FALSE )
		return -1;
	this->file.blockNum  = blockNum; // this->devCtl.ScsiGetLBA(this->capa->logicalBlockAddress);
	this->file.blockSize = (unsigned)DevCtl::ScsiGetLBA(this->capa->blockLength);
	if ( this->file.blockNum == I64(0) || this->file.blockSize == 0 )
		return -1;
#if		defined(WIN32)
	if ( this->file.handle != INVALID_HANDLE_VALUE )
		return -1;
	this->file.handle = ::CreateFile(
	    path
	  , GENERIC_READ | GENERIC_WRITE
	  , FILE_SHARE_READ
	  , NULL
	  , CREATE_NEW
	  , FILE_ATTRIBUTE_NORMAL | FILE_FLAG_WRITE_THROUGH | FILE_FLAG_NO_BUFFERING | (async ? FILE_FLAG_OVERLAPPED : 0)
	  , NULL );
	if ( this->file.handle == INVALID_HANDLE_VALUE )
		return -1;
#elif	defined(LINUX)
	if ( this->file.fd != -1 )
		return -1;
	if ( (this->file.fd = open(path, O_EXCL| O_CREAT | O_RDWR | O_SYNC)) < 0 )
	{
		this->file.fd = -1;
		return -1;
	}
#endif
	this->file.async = async;
	{
		if ( this->file.path != NULL )
			delete [] this->file.path;
		if ( (this->file.path = new char [strlen(path) + 1]) == NULL )
		{
			this->FileClose();
			return -1;
		}
		strcpy(this->file.path, path);
	}

	{
		BYTE	*buffer = NULL;
		INT64	addr = 0;
		time_t	tmNow, tmLast;

		if ( (buffer = new BYTE [this->file.blockSize]) == NULL )
			return -1;
		memset(buffer, 0, this->file.blockSize);
		time(&tmLast);
		do
		{
			time(&tmNow);
			if ( difftime(tmNow, tmLast) > 1 )
			{
				DevTest::fprintf(stderr, "Create %16llu/%-16llu (%3u%%)\r"
				  , addr * this->file.blockSize
				  , this->file.blockNum * this->file.blockSize
				  , (unsigned)( I64(100) * addr / this->file.blockNum) );
				tmLast = tmNow;
			}
			addr += (INT64)(TESTFILE_CREATESIZE / this->file.blockSize);
			if ( addr > this->file.blockNum - I64(1) )
				addr = this->file.blockNum - I64(1);
			this->FileWrite( addr, 1, this->file.blockSize, buffer);
		} while ( addr < this->file.blockNum - I64(1) );
		DevTest::fprintf(stderr, "\n");
		delete [] buffer;
	}

	return 0;
}

int		DevTest::DevParam::FileRead(INT64 addr, unsigned count, unsigned size, void * buffer)
{
#if		defined(WIN32)
	if ( this->file.handle != INVALID_HANDLE_VALUE )
	{
		LARGE_INTEGER	pos;
		DWORD			dwResult;

		if ( addr + (INT64)count > this->file.blockNum )
			return -1;
		pos.QuadPart = addr * (INT64) this->file.blockSize;
		if ( this->file.async )
		{
			OVERLAPPED		ol;

			memset(&ol, 0, sizeof(ol));
			ol.Offset     = pos.LowPart;
			ol.OffsetHigh = pos.HighPart;
			if ( ::ReadFile(this->file.handle, buffer, size, &dwResult, &ol) == FALSE )
			{
				if ( ::GetLastError() != ERROR_IO_PENDING )
					return -1;
			}
			if ( ::GetOverlappedResult(this->file.handle, &ol, &dwResult, TRUE) == 0 )
				return -1;
			dwResult = size;
		} else
		{
			if ( pos.LowPart = ::SetFilePointer(this->file.handle, pos.LowPart, &pos.HighPart, FILE_BEGIN) == INVALID_SET_FILE_POINTER
			  && ::GetLastError() != NO_ERROR )
				return -1;
			if ( ::ReadFile(this->file.handle, buffer, size, &dwResult, NULL) == FALSE )
				return -1;
		}
		return (int)dwResult;
	}
#elif	defined(LINUX)
	if ( this->file.fd != -1 )
	{
		off_t			pos;
		ssize_t			res;

		if ( addr + (INT64)count > this->file.blockNum )
			return -1;
		pos = addr * (INT64) this->file.blockSize;
		if ( lseek(this->file.fd, pos, SEEK_SET) < 0 )
			return -1;
		if ( (res = read(this->file.fd, buffer, size)) != size )
			return -1;
		return (int) res;
	}
#endif

	return -1;
}

int		DevTest::DevParam::FileWrite(INT64 addr, unsigned count, unsigned size, const void * buffer)
{
#if		defined(WIN32)
	if ( this->file.handle != INVALID_HANDLE_VALUE )
	{
		LARGE_INTEGER	pos;
		DWORD			dwResult;

		if ( addr + (INT64)count > this->file.blockNum )
			return -1;
		pos.QuadPart = addr * (INT64) this->file.blockSize;
		if ( this->file.async )
		{
			OVERLAPPED		ol;

			memset(&ol, 0, sizeof(ol));
			ol.Offset     = pos.LowPart;
			ol.OffsetHigh = pos.HighPart;
			if ( ::WriteFile(this->file.handle, buffer, size, &dwResult, &ol) == FALSE )
			{
				if ( ::GetLastError() != ERROR_IO_PENDING )
					return -1;
			}
			if ( ::GetOverlappedResult(this->file.handle, &ol, &dwResult, TRUE) == 0 )
				return -1;
			dwResult = size;
		} else
		{
			if ( pos.LowPart = ::SetFilePointer(this->file.handle, pos.LowPart, &pos.HighPart, FILE_BEGIN) == INVALID_SET_FILE_POINTER
			  && ::GetLastError() != NO_ERROR )
				return -1;
			if ( ::WriteFile(this->file.handle, buffer, size, &dwResult, NULL) == FALSE )
				return -1;
		}
		return (int)dwResult;
	}
#elif	defined(LINUX)
	if ( this->file.fd != -1 )
	{
		off_t			pos;
		ssize_t			res;

		if ( addr + (INT64)count > this->file.blockNum )
			return -1;
		pos = addr * (INT64) this->file.blockSize;
		if ( lseek(this->file.fd, pos, SEEK_SET) < 0 )
			return -1;
		if ( (res = write(this->file.fd, buffer, size)) != size )
			return -1;
		return (int) res;
	}
#endif

	return -1;
}

int		DevTest::DevParam::FileClose(void)
{
#if		defined(WIN32)
	if ( this->file.handle != INVALID_HANDLE_VALUE )
	{
		::CloseHandle(this->file.handle);
		this->file.handle = INVALID_HANDLE_VALUE;
		if ( this->file.path != NULL )
			::DeleteFile(this->file.path);
	}
#elif	defined(LINUX)
	if ( this->file.fd != -1 )
	{
		::close(this->file.fd);
		this->file.fd = -1;
		if ( this->file.path != NULL )
			::remove(this->file.path);
	}
#endif
	if ( this->file.path != NULL )
	{
		delete [] this->file.path;
		this->file.path = NULL;
	}

	return 0;
}

