/////////////////////////////////////////////////////////////////////////
//	"devtest.h"
//------------------------------------------------------------------------
//	$Id: devtest.h 165 2006-03-21 11:56:20Z taro $
//////////////////////////////////////////////////////////////////////////
#ifndef	DEVTEST_H_INCLUDED
#define	DEVTEST_H_INCLUDED

#define	DEVTEST_USE_CDROM		// Scan "Cdrom#" Device
#define	DEVTEST_USE_UDFLIB		// Check UDF Disc
#define	DEVTEST_USE_STL_LIST
#define	DEVTEST_DUMP_CSTYLE		// for Debug & Development
//#define	DEVTEST_USE_FDISK		// EXPERIMENTAL
#define	DEVTEST_USE_DISCSPEED

#if		defined(WIN32)
# include	"win32/devctl.h"
# include	"perflib.h"
#elif	defined(LINUX)
# include	"linux/devctl.h"
#endif

#if	defined(DEVTEST_USE_UDFLIB)
# include	"udflib.h"
#endif

#if	defined(DEVTEST_USE_STL_LIST)
# include	<list>
#endif

#include	<stdio.h>

#if	defined(WIN32)
# define	DEVTEST_TITLE	"***** DevTest - Device Benchmark Test Program for Windows 2000/XP *****"
#elif defined(LINUX)
# define	DEVTEST_TITLE	"***** DevTest - Device Benchmark Test Program for Linux (beta)    *****"
#endif
#define		DEVTEST_VERSION	"                                        Version 1.05a edition #02"

class	DevTest
{
public:
	enum devtest_enum
	{
		SECTOR_SIZE_MAX    = 65536,
		MAX_BUFFER_SIZE    = 65536,
		MIN_BUFFER_SIZE    = 256,
		TEST_TIME_SEC      = 3,
		RND_TEST_TIME_SEC  = 10,
		RANDOM_SCALE       = 32768,
		TRACK_POS_SCALE    = 1024,
		RWRATIO_SCALE      = 1000,
		COMMAND_TEST_COUNT = 128,
		IBM_DISK_BLOCKSIZE = 512,
		IO_QUE_DEPTH       = 8,

		COLUMN1            = 38,
		COLUMN2            = 49,
		COLUMN3            = 60,
		COLUMN_MAX         = 80,

		CD_SPEED           =  153600,	// 150KB/S (CD-DA 176.4kB/s)
		DVD_SPEED          = 1385000,

		TESTFILE_MINSIZE   = 0x100000,	// 1M bytes
		TESTFILE_AUTORATIO = 95,		// 95%
		TESTFILE_CREATESIZE= 0x100000,	// 1M

		MODE_CDB6          =  6,
		MODE_CDB10         = 10,

		SMART_ATTRIBUTE_MAX_LENGHTH = 30,
	};
	enum byteunit_enum
	{
		BYTEUNIT_210,	// 2^10
		BYTEUNIT_103,	// 10^3
	};
	enum benchflag_enum
	{
		BEN_READ        = 0x00000001,
		BEN_WRITE       = 0x00000002,
		BEN_RW          = BEN_READ | BEN_WRITE,
		BEN_RWMASK      = BEN_READ | BEN_WRITE,

		BEN_SEQUENTIAL  = 0x00000010,
		BEN_RANDOM      = 0x00000020,

		BEN_TRACK_START = 0x00000100,
		BEN_TRACK_END   = 0x00000200,
		BEN_TRACK_USER  = 0x00000400,

		BEN_BLOCK_MIN   = 0x00001000,
		BEN_BLOCK_MAX   = 0x00002000,
		BEN_BLOCK_USER  = 0x00004000,
	};
	enum cpuusage_enum
	{
		CPUUSAGE_OFF  ,
		CPUUSAGE_SELF ,
#if	defined(WIN32)
		CPUUSAGE_ALL  ,
#endif
#if	defined(LINUX)
		CPUUSAGE_SYS  ,
		CPUUSAGE_USER ,
#endif
	};
	enum command_enum
	{
		CMD_SCAN,
		CMD_SHOW,
		CMD_DUMP,
		CMD_BENCHMARK,
		CMD_SURFSCAN,
		CMD_READBLOCK ,
#if	defined(DEVTEST_USE_DISCSPEED)
		CMD_DISCSPEED,
#endif
#if	defined(DEVTEST_USE_FDISK)
		CMD_FDISK,
#endif
		CMD_MEDIUM ,
		CMD_MAX   ,
	};
#if	defined(WIN32)
	enum scanmode_enum
	{
		SCAN_OFF  ,
		SCAN_DRIVE,
		SCAN_DISK ,
		SCAN_SCSI ,
#if	defined(DEVTEST_USE_CDROM)
		SCAN_CDROM,
#endif
		SCAN_ALL  ,
	};
#endif

	typedef INT64				Clock_t;

	class	DiskInf
	{
	public:
		unsigned				blockSize;
		INT64					blockStart;
		INT64					blockNumber;
	};
	class DevParam
	{
		friend class	DevTest;
	private:
		char						*devname;
		char						*desc;
		BOOL						scsi;
		DevCtl::ScsiInqData	 		*inqData;
		DevCtl::SCSICAPACITYDATA	*capa;
		int							pathId;
		int							targetId;
		int							lun;
		BOOL						setDiskinf;
		DiskInf						diskinf;
		struct tagFile
		{
			char					*path;
#if		defined(WIN32)
			HANDLE					handle;
#elif	defined(LINUX)
			int						fd;
#endif
			unsigned				blockSize;
			INT64					blockNum;
			BOOL					async;
		} file;

	public:
		DevParam();
		virtual ~DevParam();

		int					SetDevName(const char * text);
		const char *		GetDevName(void) const { return this->devname; }
		int					SetDesc(const char * text);
		const char *		GetDesc(void) const { return this->desc; }
		const char *		GetDesc(unsigned max);

		BOOL				CheckInqData(void) const { return this->inqData != NULL; }
		DevCtl::ScsiInqData * GetInqData(void) const { return this->inqData; }
		void				SetInqData(DevCtl::ScsiInqData * pinq) { this->inqData = pinq; }
		void				ClearInqData(void) { this->inqData = NULL; }

		BOOL				CheckDiskInf(void) const { return this->setDiskinf; }
		BOOL				CheckCapacityData(void) const { return this->capa != NULL; }
		DevCtl::SCSICAPACITYDATA *	GetCapacityData(void) const { return this->capa; }
		unsigned			GetCapacityBlockLength(void) const { return (unsigned) DevCtl::ScsiGetLBA(this->capa->blockLength); }
		INT64				GetCapacityBlockAddress(void) const { return DevCtl::ScsiGetLBA(this->capa->logicalBlockAddress); }
		void				ClearCapacityData(void);
		int					SetCapacityData(DevCtl::SCSICAPACITYDATA & buf);

		void				SetPathId(unsigned id) { this->pathId = id; }
		void				SetTargetId(unsigned id) { this->targetId = id; }
		void				SetLun(unsigned lun) { this->lun = lun; }

		int					FileOpen(const char * path, INT64 blockNum, BOOL async = FALSE);
		int					FileRead(INT64 addr, unsigned count, unsigned size, void * buffer);
		int					FileWrite(INT64 addr, unsigned count, unsigned size, const void * buffer);
		int					FileClose(void);
		BOOL				IsFileMode(void) const
		{
#if		defined(WIN32)
			return this->file.handle != INVALID_HANDLE_VALUE;
#elif	defined(LINUX)
			return this->file.fd != -1;
#endif
		}
	};

	struct DevTargetParam
	{
#if	defined(WIN32)
		DevCtl::ScsiDevData 	*scsiDev;
		DevCtl::ScsiBusData 	*busData;
#endif
		DevCtl::ScsiInqData 	*inqData;
		DevCtl::ScsiInqData		inqDataWork;
		BOOL					setInqData;
	};
#if	defined(DEVTEST_USE_STL_LIST)
	typedef std::list<DevTest::DevParam *>	DevArgList;
#else

	class	DevArgList
	{
	public:
		class	Item
		{
		public:
			DevParam	*p;
			Item		*next, *prev;
		};
		class	iterator
		{
		friend class	DevArgList;
		private:
			Item			*item;
		public:
			iterator() { this->item = NULL; }

			iterator &	operator = (const iterator & x) { this->item = x.item; return *this; }
			iterator &	operator = (const void * p) { this->item = (Item *)p; return *this; }
			DevParam * 	operator * () const { return this->item->p; }
			DevParam **	operator -> () const { return &(this->item->p); }
			int			operator == (const void * ptr) const { return this->item == ptr; }
			int			operator != (const void * ptr) const { return this->item != ptr; }
			int			operator == (const iterator & it) const { return this->item == it.item; }
			int			operator != (const iterator & it) const { return this->item != it.item; }
			iterator &	operator ++ () { if ( this->item != NULL ) this->item = this->item->next; return *this; }
			iterator &	operator -- () { if ( this->item != NULL ) this->item = this->item->prev; return *this; }
			iterator	operator ++ (int) { iterator it = *this; ++(*this); return it; }
			iterator	operator -- (int) { iterator it = *this; --(*this); return it; }
		};
	private:
		unsigned		numItem;
		iterator		itop, ilast;
		iterator		iend, icur;
	public:
		DevArgList();
		~DevArgList();

		void		clear(void);
		iterator 	insert(iterator & base, const DevParam * s);
		iterator 	push_back(const DevParam * x) { iterator base; base.item = NULL; return this->insert(base, x); }
		void		remove(iterator & x);
		iterator 	begin(void) const { return this->itop; }
		iterator 	end(void) const { return this->iend; }
		unsigned	size(void) const { return this->numItem; }
	};
#endif
	typedef DevArgList::iterator			DevArgIterator;

	struct tagOption
	{
		int			cmd;
		int			scanMode;
		unsigned	byteUnit;
		BOOL		printPartition;
		BOOL		printDevInfo;
		BOOL		printDetail;
		BOOL		dumpModeSense;
		BOOL		writeEnable;
		BOOL		benchCommand;
		BOOL		benchSeek;
		BOOL		benchRead;
		BOOL		benchWrite;
		BOOL		benchRandomRW;
		unsigned	benchRwRatio;
		const char *benchCsvResult;
		unsigned	measureTimeSec;
		INT64		addr;
		INT64		count;
		INT64		readAddr, writeAddr;
		BOOL		ignoreError;
		BOOL		zeroFill;
		int			retry;
		BOOL		scsiCmd;
		BOOL		verify;
		unsigned	defBlockSize;
		unsigned	mode_cdb;
#if	defined(DEVTEST_DUMP_CSTYLE)
		BOOL		dumpCStyle;
#endif
		const char	*filename;
		unsigned	cpuUsage;
#if	defined(DEVTEST_USE_DISCSPEED)
		unsigned	discSpeed;	// kB/s
#endif
		const char	*testPath;
		INT64		testFileSize;
		BOOL		async;		// Asynchronous
		BOOL		nobuffer;
		BOOL		smart;
	} opt;

	DevCtl		devCtl;
	typedef int (DevTest::* ReadFunc)(DevParam & devParam, INT64 addr, unsigned count, unsigned size, void * buffer);
	typedef int (DevTest::* WriteFunc)(DevParam & devParam, INT64 addr, unsigned count, unsigned size, const void * buffer);
	struct tagBen
	{
		unsigned	mode;
		unsigned	maxBufferSize;
		unsigned	testTimeSec;
		unsigned	randomRWTimeSec;
		unsigned	trackPos;		// 0 - TRACK_POS_SCALE-1
		unsigned	blockSize;
		unsigned	seekSector;		// sector/track for Seek test
		INT64		discSpeed;
		FILE		*csvResult;
		ReadFunc	read_func;
		WriteFunc	write_func;
	} ben;
	DevParam		*srcDevParam;
	DevParam		*dstDevParam;

	struct tagTestPath
	{
#if		defined(WIN32)
		HANDLE			handle;
#elif	defined(LINUX)
		int				fd;
#endif
		unsigned		blockSize;
		INT64			blockNum;
	} testpath;

	struct tagAsync
	{
#if		defined(WIN32)
		HANDLE			handle;
#endif
		int				maxQueDepth;
		int				queNo;
		struct IOCmd
		{
			void *		buffer;
			BOOL		exec;
#if		defined(WIN32)
			HANDLE		hEvent;
			OVERLAPPED	ol;
			DevCtl::SPT_DIRECT_BUFFER	spt;
#endif
		};
		IOCmd *			iotbl;
	} async;

	enum	ScsiPreventAllow
	{
		SCSI_PREVENTALLOW_UNLOCK = 0,
		SCSI_PREVENTALLOW_LOCK   = 1,
		SCSI_PREVENTALLOW_PERSISTENT_ALLOW   = 2,
		SCSI_PREVENTALLOW_PERSISTENT_PREVENT = 3,
	};
	enum	ScsiStartStop
	{
		SCSI_STARTSTOPUNIT_SPINDOWN = 0,
		SCSI_STARTSTOPUNIT_SPINUP   = 1,
		SCSI_STARTSTOPUNIT_UNLOAD   = 2,
		SCSI_STARTSTOPUNIT_LOAD     = 3,
	};
	enum	ScsiLoadUnload
	{
		SCSI_LOADUNLOAD_ABORT    = 0,
		SCSI_LOADUNLOAD_UNLOAD   = 2,
		SCSI_LOADUNLOAD_LOAD     = 3,
	};
	struct medium_tag
	{
		BOOL				setPreventAllow;
		unsigned			preventallow;

		BOOL				setStartStop;
		ScsiStartStop		startstop;

		BOOL				setLoadUnload;
		ScsiLoadUnload		loadunload;
	} medium;

#if	defined(WIN32)
	PerfLib				perflib;
	PerfLib::Counter	counter;
#endif

public:
	DevTest();
	~DevTest();

	int		ScanDevice(DevParam & devParam);
	int		ScanDeviceDrive(DevParam & devParam);
	int		ScanDeviceDisk(DevParam & devParam);
	int		ScanDeviceScsi(DevParam & devParam);
#if	defined(DEVTEST_USE_CDROM)
	int		ScanDeviceCdrom(DevParam & devParam);
#endif
	int		ScanDeviceCheck(DevParam & devParam);
	int		OpenTargetDevice(DevCtl & devCtl, DevTest::DevTargetParam & tg, DevTest::DevParam & devParam, BOOL async);
	void	CloseTargetDevice(DevCtl & devCtl, DevTest::DevTargetParam & tg, DevTest::DevParam & devParam);
	int		TargetDevice(DevTest::DevParam & devParam);
	int		PrintDevice(DevParam & devParam);
	int		PrintPartition(DevParam & devParam);
	int		PrintIbmPartition(DevParam & devParam, int & ext, INT64 & extAddr, INT64 baseAddr, BYTE * buffer);
	int		PrintFatInfo(DevTest::DevParam & devParam, INT64 addr);
	int		PrintNtfsInfo(DevTest::DevParam & devParam, INT64 addr);
	int		PrintCdrom(DevParam & devParam);
	int		PrintDump(DevParam & devParam);
	int		SurfaceScan(DevParam & devParam);
	int		ReadWriteBlock(DevTest::DevParam & devParam);
	int		SetDiscSpeed(DevParam & devParam);
#if	defined(DEVTEST_USE_FDISK)
	int		CreatePartition(DevParam & devParam);
	static void	LBA2CHS(INT64 lba, INT64 cylinder,int head, int sector, BYTE chs[3]);
	static void	GenDiskGeometory(INT64 blkNum, unsigned blkLen, INT64 & Cylinders, unsigned & TracksPerCylinder, unsigned & SectorsPerTrack);
#endif

	int		PrintDevInfo(DevParam & devParam);
	int		PrintDevInfo_vpd(DevParam & devParam);
	int		PrintDevInfo_logSense(DevTest::DevParam & devParam);
	int		PrintDevInfo_modeSense(DevParam & devParam);
	int		PrintPagecode2A(DevTest::DevParam & devParam, unsigned size, void * buffer);
	int		PrintDevInfo_defectData(DevTest::DevParam & devParam);
	int		PrintDevInfo_getConfiguration(DevTest::DevParam & devParam);
	static char *	GetFeatureCodeName(char * buffer, unsigned fcode);
	static char *	GetProfileDefinitions(char * buffer, unsigned profileNumber);
	int		PrintDevInfo_dvdStructure(DevTest::DevParam & devParam);
	int		DumpParam(unsigned addr, unsigned off, unsigned bufsiz, BYTE * buffer);

	int		BlockCopy(DevTest::DevParam & srcDev, DevTest::DevParam & dstParam);

	int		BenchReadWrite(DevParam & devParam, unsigned benmode);
	int		BenchRandomReadWrite(DevTest::DevParam & devParam, unsigned rwratio);
	int		BenchCommand(DevParam & devParam, unsigned benmode);
	int		BenchSeek(DevParam & devParam, unsigned benmode);
	int		BenchSetReadWrite(void);
	int		BenchRead(DevParam & devParam, INT64 addr, unsigned count, unsigned size, void * buffer);
	int		BenchWrite(DevParam & devParam, INT64 addr, unsigned count, unsigned size, const void * buffer);
private:
	int		read_scsi_sync(DevParam & devParam, INT64 addr, unsigned count, unsigned size, void * buffer);
	int		write_scsi_sync(DevParam & devParam, INT64 addr, unsigned count, unsigned size, const void * buffer);
	int		read_scsi_async(DevParam & devParam, INT64 addr, unsigned count, unsigned size, void * buffer);
	int		write_scsi_async(DevParam & devParam, INT64 addr, unsigned count, unsigned size, const void * buffer);
	int		read_dev_sync(DevParam & devParam, INT64 addr, unsigned count, unsigned size, void * buffer);
	int		write_dev_sync(DevParam & devParam, INT64 addr, unsigned count, unsigned size, const void * buffer);
	int		read_dev_async(DevParam & devParam, INT64 addr, unsigned count, unsigned size, void * buffer);
	int		write_dev_async(DevParam & devParam, INT64 addr, unsigned count, unsigned size, const void * buffer);
	int		read_file_sync(DevParam & devParam, INT64 addr, unsigned count, unsigned size, void * buffer);
	int		write_file_sync(DevParam & devParam, INT64 addr, unsigned count, unsigned size, const void * buffer);
	int		read_file_async(DevParam & devParam, INT64 addr, unsigned count, unsigned size, void * buffer);
	int		write_file_async(DevParam & devParam, INT64 addr, unsigned count, unsigned size, const void * buffer);

public:
	int			GetDiskInf(DevCtl & devCtl, DevTest::DevParam & devParam, DevTest::DiskInf & diskinf);
	int			ReadDev(DevCtl & devCtl, DevTest::DevParam & devParam, INT64 addr, unsigned count, unsigned size, void * buffer);
	INT64		GetTestFileSize(void);
	static int	TestFileOpen(DevCtl & devCtl, DevTest::DevParam & devParam, const char * path, INT64 size, BOOL async = FALSE);
	static void	TestFileClose(DevCtl & devCtl, DevTest::DevParam & devParam);
	BOOL	IsAsyncMode(void) { return this->opt.async && this->async.maxQueDepth > 0; }
#if	defined(WIN32)
	int		AsyncInit(HANDLE handle);
	int		AsyncRead(HANDLE handle, INT64 addr, unsigned count, unsigned size, void * buffer);
	int		AsyncWrite(HANDLE handle, INT64 addr, unsigned count, unsigned size, const void * buffer);
	int		AsyncWait(void);
	int		AsyncTerm(void);
#else
	int		AsyncWait(void) { return 0; }
	int		AsyncTerm(void) { return 0; }
#endif

	static unsigned	SC32(const BYTE * p) { return DevCtl::ScsiUint32(p); }
	static unsigned	SC24(const BYTE * p) { return DevCtl::ScsiUint24(p); }
	static unsigned	SC16(const BYTE * p) { return DevCtl::ScsiUint16(p); }
	static const char *	GetOnOff(BOOL flag) { return flag ? "ON" : "OFF"; }
	static int		fprintf(FILE * fp, const char * form, ... );
	static char *	Fmt(char * buffer, const char * form, ... );
	static void		PutFmt(int spc, const char * text1, const char * val, const char *text2);
	static void		PutPos(int pos, ... );
	static int		GetNonZeroSize(unsigned size, BYTE * buffer);
	void			PrintDumpOct(const char * msg, int ascii, unsigned size, BYTE * buffer);
	void			PrintDumpHex(unsigned addr, unsigned blockSize, unsigned size, BYTE * buffer);
#if	defined(DEVTEST_DUMP_CSTYLE)
	void			PrintDumpCStyle(unsigned size, BYTE * buffer);
#endif
	char *			FormatSize(char * buffer, INT64 size);
	char *			FormatTimeRemain(char * buffer, unsigned sec);
	char *			FormatTime(char * buffer, unsigned sec);
	static INT64	GetNumberSize(const char *s, unsigned umode);
#if	defined(WIN32)
	static char *	GetErrMsg(int errcode);
#endif
	char *			GetPartitionType(unsigned type);
	Clock_t			Clock(void);

	//////////////////////////////////////////////////////////////////////
	//	SMART
public:
	static UINT64		ATA_GET48(void *p)	{ return (
		 ((UINT64)((BYTE *)(p))[0])
		+((UINT64)((BYTE *)(p))[1]<< 8)
		+((UINT64)((BYTE *)(p))[2]<<16)
		+((UINT64)((BYTE *)(p))[3]<<24)
		+((UINT64)((BYTE *)(p))[4]<<32)
		+((UINT64)((BYTE *)(p))[5]<<40)
	); }
	static DWORD		ATA_GET32(void *p)	{ return (
		 ((DWORD)((BYTE *)(p))[0])
		+((DWORD)((BYTE *)(p))[1]<< 8)
		+((DWORD)((BYTE *)(p))[2]<<16)
		+((DWORD)((BYTE *)(p))[3]<<24)
	); }
	static unsigned		ATA_GET16(void *p)	{ return (
		 ((unsigned)((BYTE *)(p))[0])
		+((unsigned)((BYTE *)(p))[1]<< 8)
	); }
	enum	SmartVendor
	{
		SMART_VENDOR_GENERAL,
		SMART_VENDOR_FUJITSU,
		SMART_VENDOR_HGST   ,
		SMART_VENDOR_IBM    ,
		SMART_VENDOR_MAXTOR ,
		SMART_VENDOR_SAMSUNG,
		SMART_VENDOR_SEAGATE,
		SMART_VENDOR_TOSHIBA,
		SMART_VENDOR_WD     ,
	};
	enum	SmartAttrRawType
	{
		SMART_RAW_NORMAL ,
		SMART_RAW_DUMPHEX,
		SMART_RAW_DIVWORD,
	};
	struct SmartAttrItem
	{
		unsigned			id;
		SmartAttrRawType	type;
		char				*desc;
	};
	typedef std::list<SmartAttrItem *>	SmartAttrList;
	struct smart_tag
	{
		SmartAttrList	attrList;
	} smart;

	int						PrintSmartInfo(DevTest::DevParam & devParam);
	static TCHAR * 			FormatAtaString(TCHAR * out, const DevCtl::ATA_WORD * inp, int n);
	static DevTest::SmartVendor	GetAtaVendor(DevCtl::AtaIdentifyDevice & identify);
	static UINT64			GetAtaCapacity(DevCtl::AtaIdentifyDevice & identify);
	static const TCHAR *	GetSmartAttributeIdName(DevTest::SmartVendor vendor, unsigned id, unsigned length = 0, TCHAR * buffer = NULL);
	SmartAttrItem *			GetSmartAttrItem(unsigned id);
	const TCHAR * 			GetSmartRawValue(DevCtl::AtaSmartAttribute & attr, unsigned length, TCHAR * buffer);

	enum	DispType
	{
		DISPTYPE_SHORT_TEXT    ,
		DISPTYPE_LBA64_CAPACITY,
		DISPTYPE_UINT32_512SIZE,
		DISPTYPE_UINT16_BITS   ,
		DISPTYPE_UINT16_BITDEF ,
	};
	struct DispParamBitDef
	{
		unsigned	bit;
		int			set;
		const TCHAR	*text;
	};
	struct DispParam
	{
		DispType			disptype;
		const TCHAR			*name;
		union data_tag
		{
		void				*data;
		const TCHAR			*text;
		WORD				*u16;
		DWORD	 			*u32;
		UINT64				*u64;
		} data;
		unsigned			tblnum;
		union tbl_tag
		{
		void			*tbl;
		DispParamBitDef	*bitdef;
		} tbl;
	};
	int			DisplayInfo(unsigned tblnum, DevTest::DispParam * tbl);

#if	defined(DEVTEST_USE_UDFLIB)
	//////////////////////////////////////////////////////////////////////
	//	UDF
public:
	UdfLib				udflib;

	int			PrintDisc(DevParam & devParam);
#endif

	//////////////////////////////////////////////////////////////////////
	//	Linux ext2
public:
	int			PrintExt2Info(DevParam & devParam, INT64 addr, INT64 addrMax);

	//////////////////////////////////////////////////////////////////////
	//	CPU USAGE
public:
	int			InitCpuUsage(void);
	void		CollectCpuUsage(void);
	double		GetCpuUsage(void);
	void		TermCpuUsage(void);
};

#endif
