/////////////////////////////////////////////////////////////////////////
//	"devtest_lib.cpp"
//------------------------------------------------------------------------
//	$Id: devtest_lib.cpp 166 2006-03-21 12:03:00Z taro $
//////////////////////////////////////////////////////////////////////////

#include	<stdlib.h>
#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	<ctype.h>
#include	<time.h>
#if	defined(LINUX)
# include	<sys/time.h>
# include	<sys/resource.h>
# include	<unistd.h>
#endif
#include	"devtest.h"
#include	"defs.h"
#include	"tform.h"

#if	defined(WIN32)
# include	<crtdbg.h>
#elif defined(LINUX)
# include	<assert.h>
# define	_ASSERT
#endif

#if defined(_DEBUG) && defined(WIN32)
# define	DBGREPORT(param_)	_CrtDbgReport param_
#else
# define	DBGREPORT(param_)
#endif

int		DevTest::GetDiskInf(DevCtl & devCtl, DevTest::DevParam & devParam, DevTest::DiskInf & diskinf)
{
#if	defined(WIN32)
	unsigned					bsize_geo, bsize_toc, bsize_scsi;
	UINT64						len_part, len_info, len_toc, len_scsi;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "DevTest::GetDiskInf()\n"));
	bsize_geo = bsize_scsi = 0;
	len_part = len_info = len_toc = len_scsi = I64(0);
	{
		DevCtl::SCSICAPACITYDATA	capaBuf;

		if ( devParam.CheckInqData() != FALSE
		  && devCtl.ScsiReadCapacity(devParam.pathId, devParam.targetId, devParam.lun, sizeof(capaBuf), &capaBuf ) == 0 )
		{
			bsize_scsi = (unsigned) devCtl.ScsiGetLBA(capaBuf.blockLength        );
			len_scsi   =            devCtl.ScsiGetLBA(capaBuf.logicalBlockAddress) * bsize_scsi;
		}
	}

	{
		DISK_GEOMETRY				geo;

		memset(&geo, 0, sizeof(geo));
		if ( devCtl.DiskGetDriveGeometry(geo)  == 0 
		  || devCtl.StorageGetMediaTypes(geo)  == 0 
		  || devCtl.CdromGetDriveGeometry(geo) == 0
		  || geo.BytesPerSector > 0 )
		{
			bsize_geo = geo.BytesPerSector;
		}
	}
	{
		PARTITION_INFORMATION		partInfo;

		memset(&partInfo  , 0, sizeof(partInfo  ));
		if ( devCtl.DiskGetPartitionInfo(partInfo) == 0 && partInfo.PartitionLength.QuadPart > 0 )
		{
			len_info = partInfo.PartitionLength.QuadPart;
		//	diskinf.blockStart = partInfo.StartingOffset.QuadPart;
		}
	}

	{
		DevCtl::SCSITOCDATA			toc;

		if ( devCtl.CdromReadToc(toc) == 0 )
		{
			unsigned	toc_len;

			toc_len = SCSI_GET16(toc.tocLength);
			if ( toc_len >= 10 )
			{
				int			n;

				diskinf.blockStart = 0;
				for ( n = 0; n < sizeof(toc.track)/sizeof(*toc.track); n++ )
				{
					if ( n * sizeof(*toc.track) + 10 >= toc_len )
						break;
					if ( toc.track[n].no == 0xAA )
						break;
					if ( toc.track[n].adrctl & 0x10 && toc.track[n].adrctl & 0x04 )
					{	// DATA TRACK
						INT64	adr, nextadr;
#if 1
						unsigned	ff, ss, mm;

						mm = toc.track[n].cdAddr[1];
						ss = toc.track[n].cdAddr[2];
						ff = toc.track[n].cdAddr[3];
						adr = (INT64)mm * I64(60)*I64(75) + (INT64)ss*I64(75) + (INT64)ff;
						mm = toc.track[n+1].cdAddr[1];
						ss = toc.track[n+1].cdAddr[2];
						ff = toc.track[n+1].cdAddr[3];
						nextadr = (INT64)mm * I64(60)*I64(75) + (INT64)ss*I64(75) + (INT64)ff;
#else
						adr     = (INT64) SCSI_GET32(toc.track[n  ].cdAddr);
						nextadr = (INT64) SCSI_GET32(toc.track[n+1].cdAddr);
#endif
					//	fprintf(stderr, "Track addr %u, next addr %u\n", (unsigned) adr, (unsigned)nextadr);
						if ( bsize_scsi > 0 )
							bsize_toc = bsize_scsi;
						else if ( bsize_geo > 0 )
							bsize_toc = bsize_geo;
						else
							bsize_toc = 2048;
						len_toc += (nextadr - adr) * bsize_toc;
					}
				}
			}
		}
	}

	{
		GET_LENGTH_INFORMATION		lengthInfo;

		memset(&lengthInfo, 0, sizeof(lengthInfo));
		if ( devCtl.DiskGetLengthInfo(lengthInfo) == 0 && lengthInfo.Length.QuadPart > 0 )
			len_info = lengthInfo.Length.QuadPart;
	}
	if ( len_info == 0 && len_part == 0 && len_toc == 0 && len_scsi == 0 )
		return -1;
	diskinf.blockSize = 0;
	diskinf.blockNumber = I64(0);
	diskinf.blockStart = I64(0);
	switch ( devCtl.GetHandleType() )
	{
	case DevCtl::HT_DISK:	// physical
		if ( this->opt.scsiCmd && bsize_scsi > 0 )
		{
			DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  Set SCSI Capacity\n"));
			diskinf.blockSize   = bsize_scsi;
			diskinf.blockNumber = len_scsi / diskinf.blockSize;
		}
		break;
	case DevCtl::HT_DRIVE:	// drive-letter
		if ( this->opt.scsiCmd == FALSE && len_info > 0 && bsize_geo > 0 )
		{
			DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  Set LengthInfo\n"));
			diskinf.blockSize   = bsize_geo;
			diskinf.blockNumber = len_info / diskinf.blockSize;
		} else if ( this->opt.scsiCmd == FALSE && len_part > 0 && bsize_geo > 0 )
		{
			DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  Set PartitionInfo\n"));
			diskinf.blockSize   = bsize_geo;
			diskinf.blockNumber = len_part / diskinf.blockSize;
		}
		break;
	case DevCtl::HT_SCSI:	// scsi
		if ( bsize_scsi > 0 )
		{
			DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  Set SCSI Capacity\n"));
			diskinf.blockSize   = bsize_scsi;
			diskinf.blockNumber = len_scsi / diskinf.blockSize;
		}
		break;
	case DevCtl::HT_UNKNOWN:
	default:
		break;
	}
	if ( diskinf.blockSize == 0 && diskinf.blockNumber == 0 )
	{
		if ( this->opt.scsiCmd && bsize_scsi > 0 )
			diskinf.blockSize = bsize_scsi;
		else if ( bsize_geo > 0 )
			diskinf.blockSize = bsize_geo;
		else if ( bsize_scsi > 0 )
			diskinf.blockSize = bsize_scsi;
		else
			return -1;

		if ( this->opt.scsiCmd && len_scsi > 0 )
			diskinf.blockNumber = len_scsi / diskinf.blockSize;
		else if ( len_info > 0 )
			diskinf.blockNumber = len_info / diskinf.blockSize;
		else if ( len_part > 0 )
			diskinf.blockNumber = len_part / diskinf.blockSize;
		else if ( len_toc > 0 )
			diskinf.blockNumber = len_toc / diskinf.blockSize;
		else if ( len_scsi > 0 )
			diskinf.blockNumber = len_toc / diskinf.blockSize;
		else
			return -1;
	}

#elif	defined(LINUX)

	{
		DevCtl::SCSICAPACITYDATA	capaBuf;

		if ( devParam.CheckInqData() == FALSE )
		{
		//	fprintf(stderr, "scsi error\n");
			return -1;
		}
		if ( devCtl.ScsiReadCapacity(devParam.pathId, devParam.targetId, devParam.lun, sizeof(capaBuf), &capaBuf ) < 0 )
			return -1;
		diskinf.blockSize   = (unsigned) devCtl.ScsiGetLBA(capaBuf.blockLength        );
		diskinf.blockNumber =            devCtl.ScsiGetLBA(capaBuf.logicalBlockAddress);
		diskinf.blockStart  = 0;
	}
#endif
	return 0;
}

int		DevTest::ReadDev(DevCtl & devCtl, DevTest::DevParam & devParam, INT64 addr, unsigned count, unsigned size, void * buffer)
{
#if	defined(WIN32)
	if ( this->opt.scsiCmd )
	{
#endif
		if ( devParam.CheckInqData() == FALSE )
			return -1;
		return devCtl.ScsiRead(devParam.pathId, devParam.targetId, devParam.lun, addr, count, size, buffer );
#if	defined(WIN32)
	} else
	{
		INT64			pos;
		DWORD			dwRes;

		if ( devParam.setDiskinf == FALSE )
		{
			DevTest::DiskInf	diskinf;

			if ( this->GetDiskInf(devCtl, devParam, diskinf) < 0 )
				return -1;
			devParam.diskinf = diskinf;
			devParam.setDiskinf = TRUE;
		}
		pos = addr * (INT64)devParam.diskinf.blockSize;
		if ( devCtl.SetFilePointer(pos, FILE_BEGIN) < 0 )
			return -1;
		if ( ::ReadFile( devCtl.GetHandle(), buffer, size, &dwRes, NULL) == 0 )
			return -1;
		return (int)dwRes;
	}
#endif
}


INT64		DevTest::GetTestFileSize(void)
{
	INT64	testSize;

	if ( (testSize = this->opt.testFileSize) == I64(0) )
	{
#if	defined(WIN32)
		ULARGE_INTEGER	free_bytes, total_bytes, total_free_bytes;
		char			path[512], *fn;

		if ( ::GetFullPathName(this->opt.testPath, sizeof(path), path,  &fn) == 0 )
			return -1;
#if	defined(DEBUG)
		fprintf(stderr, "full path = %s, file = %s\n", path, fn);
#endif
		if ( fn != NULL )
			*fn = 0;
		if ( ::GetDiskFreeSpaceEx(path, &free_bytes, &total_bytes, &total_free_bytes) == 0 )
			return -1;
#if	defined(DEBUG)
		fprintf(stderr, "%s, %"FI64"u, %"FI64"u, %"FI64"u\n", path, free_bytes.QuadPart, total_bytes.QuadPart, total_free_bytes.QuadPart);
#endif
		testSize = free_bytes.QuadPart * (INT64)TESTFILE_AUTORATIO / I64(100);	// 95%
#endif
	}
	return testSize;
}

int		DevTest::TestFileOpen(DevCtl & devCtl, DevTest::DevParam & devParam, const char * path, INT64 size, BOOL async) // static
{
	unsigned	blockSize;
	INT64		blockNum;
	INT64		testBlockNum;

#if	defined(DEBUG)
	fprintf(stderr, "WriteTsetFileOpen %s, size = %"FI64"u\n", path, size);
#endif
	if ( devParam.CheckCapacityData() == FALSE )
		return -1;
	blockNum  =           devCtl.ScsiGetLBA(devParam.GetCapacityData()->logicalBlockAddress);
	blockSize = (unsigned)devCtl.ScsiGetLBA(devParam.GetCapacityData()->blockLength);
	testBlockNum = size / (INT64)blockSize;

#if	defined(DEBUG)
	fprintf(stderr, "  blockSize %u, blockNum %"FI64"d, testBlock %"FI64"d \n"
	  , (unsigned)blockSize
	  , (INT64)blockNum
	  , (INT64)testBlockNum);
#endif
	if ( size < (INT64)TESTFILE_MINSIZE || testBlockNum > blockNum )	// <= 1M error
		return -1;
	if  ( devParam.FileOpen(path, testBlockNum, async) < 0 )
	{
		fprintf(stderr, "TestFile open error! (%s)\n", path);
		return -1;
	}
	return 0;
}

void	DevTest::TestFileClose(DevCtl & devCtl, DevTest::DevParam & devParam) // static
{
#if	defined(DEBUG)
	fprintf(stderr, "WriteTsetFileClose\n");
#endif
	devParam.FileClose();
}

#if 0
int		DevTest::TestRead(DevCtl & devCtl, DevTest::DevParam & devParam, INT64 addr, unsigned count, unsigned size, void * buffer)
{
	if ( this->opt.async )
	{
#if		defined(WIN32)
		if ( this->opt.testPath != NULL  && devParam.file.handle != INVALID_HANDLE_VALUE )
		{
			if ( this->AsyncRead(devParam.file.handle, addr, count, size, buffer) < 0 )
				return -1;
			return size;	// no error
		} else
		{
			int		ret;

			if ( this->async.iotbl == NULL )
			{
				if ( this->AsyncInit(devCtl.GetHandle()) < 0 )
					return -1;
			}
			if ( this->async.queNo >= this->async.maxQueDepth )
			{
				this->AsyncWait();
				this->async.queNo = 0;
			}
			memset(&this->async.iotbl[this->async.queNo].ol, 0, sizeof(this->async.iotbl[this->async.queNo].ol));
			this->async.iotbl[this->async.queNo].ol.hEvent  = this->async.iotbl[this->async.queNo].hEvent;
			this->async.iotbl[this->async.queNo].exec = TRUE;
			::ResetEvent(this->async.iotbl[this->async.queNo].hEvent);	// nonsignaled
			ret = devCtl.ScsiRead(
			  devParam.inqData->GetPathId()
			, devParam.inqData->GetTargetId()
			, devParam.inqData->GetLun()
			, addr, count, size, this->async.iotbl[this->async.queNo].buffer
			, this->async.iotbl[this->async.queNo].spt, this->async.iotbl[this->async.queNo].ol);
			this->async.queNo++;

			if ( ::GetLastError() == ERROR_IO_PENDING )
			{
				DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  ERROR_IO_PENDING in TestRead() : queNo = %2d\n", this->async.queNo));
			}

			return size;	// no error
		}
#elif	defined(LINUX)
		return -1;	// error
#endif
	} else
	{
#if		defined(WIN32)
		if ( this->opt.testPath != NULL  && devParam.file.handle != INVALID_HANDLE_VALUE )
		{
			return devParam.FileRead(addr, count, size, buffer);
		}
#elif	defined(LINUX)
		if ( this->opt.testPath != NULL  && devParam.file.fd != -1 )
		{
			return devParam.FileRead(addr, count, size, buffer);
		}
#endif
		else
		{
			return devCtl.ScsiRead(
			  devParam.inqData->GetPathId()
			, devParam.inqData->GetTargetId()
			, devParam.inqData->GetLun()
			, addr, count, size, buffer);
		}
	}
	return -1;
}
int		DevTest::TestWrite(DevCtl & devCtl, DevTest::DevParam & devParam, INT64 addr, unsigned count, unsigned size, const void * buffer)
{
	if ( this->opt.async )
	{
#if		defined(WIN32)
		if ( this->opt.testPath != NULL  && devParam.file.handle != INVALID_HANDLE_VALUE )
		{
			if ( this->AsyncWrite(devParam.file.handle, addr, count, size, buffer) < 0 )
				return -1;
			return size;	// no error
		} else
		{
#if	defined(DEVTEST_USE_WRITETEST)
		//	memcpy(this->async.iotbl[this->async.queNo].buffer, buffer, size);
#endif
		}
#elif	defined(LINUX)
		return -1;	// error
#endif
	} else
	{
#if		defined(WIN32)
		if ( this->opt.testPath != NULL  && devParam.file.handle != INVALID_HANDLE_VALUE )
		{
			return devParam.FileWrite(addr, count, size, buffer);
		}
#elif	defined(LINUX)
		if ( this->opt.testPath != NULL  && devParam.file.fd != -1 )
		{
			return devParam.FileWrite(addr, count, size, buffer);
		}
#endif
		else
		{
			if ( this->opt.writeEnable == FALSE )
				return -1;
#if	defined(DEVTEST_USE_WRITETEST)
#endif
		}
	}
	return -1;
}
#endif

#if	defined(WIN32)
int		DevTest::AsyncInit(HANDLE handle)
{
	int				i;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "AsyncInit(handle=0x%08X) : maxQueDepth=%2d\n", handle, this->async.maxQueDepth));
	if ( this->opt.async == FALSE || this->async.maxQueDepth < 1 )
		return -1;
	if ( (this->async.iotbl = new DevTest::tagAsync::IOCmd [this->async.maxQueDepth]) == NULL )
		return -1;
	for ( i = 0; i < this->async.maxQueDepth; i++ )
	{
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "    init no.%02d\n", i));
		this->async.iotbl[i].exec = FALSE;
		if ( (this->async.iotbl[i].buffer = ::VirtualAlloc(NULL, this->ben.maxBufferSize, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE)) == NULL )
			return -1;
		if ( (this->async.iotbl[i].hEvent = ::CreateEvent(NULL, TRUE, FALSE, NULL)) == NULL )	// manual-reset, nonsignaled
			return -1;
	//	::ResetEvent(this->async.iotbl[i].ol.hEvent);
	}
	this->async.queNo = 0;
	this->async.handle = handle;

	return 0;
}

int		DevTest::AsyncRead(HANDLE handle, INT64 addr, unsigned count, unsigned size, void * buffer)
{
	LARGE_INTEGER	pos;
	DWORD			dwResult;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "AsynRead(handle=0x%08X, addr=%I64u, count=%u, size=%u) : queNo=%d\n"
	  , handle, addr, count, size, this->async.queNo));
	if ( this->async.iotbl == NULL )
	{
		if ( this->AsyncInit(handle) < 0 )
			return -1;
	}
	if ( this->async.queNo >= this->async.maxQueDepth )
	{
		this->AsyncWait();
		this->async.queNo = 0;
	}

	pos.QuadPart = addr * (INT64)(size/count);
	memset(&this->async.iotbl[this->async.queNo].ol, 0, sizeof(this->async.iotbl[this->async.queNo].ol));
	this->async.iotbl[this->async.queNo].ol.Offset     = pos.LowPart;
	this->async.iotbl[this->async.queNo].ol.OffsetHigh = pos.HighPart;
	this->async.iotbl[this->async.queNo].ol.hEvent     = this->async.iotbl[this->async.queNo].hEvent;
	this->async.iotbl[this->async.queNo].exec          = TRUE;
	::ResetEvent(this->async.iotbl[this->async.queNo].hEvent);	// nonsignaled
	if ( ::ReadFile(handle
	       , this->async.iotbl[this->async.queNo].buffer
	       , size
	       , &dwResult
	       , &this->async.iotbl[this->async.queNo].ol) == FALSE )
	{
		if ( ::GetLastError() == ERROR_IO_PENDING )
		{
			DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  ERROR_IO_PENDING in AsyncRead() : queNo = %2d\n", this->async.queNo));
		}
	}
	this->async.queNo++;

	return 0;
}

int		DevTest::AsyncWrite(HANDLE handle, INT64 addr, unsigned count, unsigned size, const void * buffer)
{
	LARGE_INTEGER	pos;
	DWORD			dwResult;

	if ( this->async.iotbl == NULL )
	{
		if ( this->AsyncInit(handle) < 0 )
			return -1;
	}
	if ( this->async.queNo >= this->async.maxQueDepth )
	{
		this->AsyncWait();
		this->async.queNo = 0;
	}
	pos.QuadPart = addr * (INT64)(size/count);
	memset(&this->async.iotbl[this->async.queNo].ol, 0, sizeof(this->async.iotbl[this->async.queNo].ol));
	this->async.iotbl[this->async.queNo].ol.Offset     = pos.LowPart;
	this->async.iotbl[this->async.queNo].ol.OffsetHigh = pos.HighPart;
	this->async.iotbl[this->async.queNo].ol.hEvent     = this->async.iotbl[this->async.queNo].hEvent;
	this->async.iotbl[this->async.queNo].exec          = TRUE;
//	memcpy(this->async.iotbl[this->async.queNo].buffer, buffer, size);
	::ResetEvent(this->async.iotbl[this->async.queNo].hEvent);	// nonsignaled
	if ( ::WriteFile(handle
	       , this->async.iotbl[this->async.queNo].buffer
	       , size
	       , &dwResult
	       , &this->async.iotbl[this->async.queNo].ol) == FALSE )
	{
		if ( ::GetLastError() == ERROR_IO_PENDING )
		{
			DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  ERROR_IO_PENDING in AsyncWrite() : queNo = %2d\n", this->async.queNo));
		}
	}
//	 else
//		this->async.iotbl[this->async.queNo].exec      = FALSE;
	this->async.queNo++;

	return 0;
}

int		DevTest::AsyncWait(void)
{
	int				i, err = 0;

	if ( this->async.iotbl == NULL )
		return 0;
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "AsyncWait() : handle=0x%08X, queNo=%d\n", this->async.handle, this->async.queNo));
	for ( i = 0; i < this->async.maxQueDepth; i++ )
	{
		if ( this->async.iotbl[i].exec )
		{
			DWORD		dw;
			int			res;

			DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  Call GetOverlappedResult() in AsyncWait() : queNo = %2d\n", i));
			res = ::GetOverlappedResult(this->async.handle, &this->async.iotbl[i].ol, &dw, TRUE);
#if	defined(_DEBUG)
			if ( ! res )	// error
			{
				DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  GetOverlappedResult() Error in AsyncWait() : queNo = %2d\n", i));
			}
#endif
			this->async.iotbl[i].exec = FALSE;
		}
	}
	this->async.queNo = 0;

	return err > 0 ? -1 : 0;
}

int		DevTest::AsyncTerm(void)
{
	int				i;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "AsyncTerm()\n"));
	if ( this->opt.async == FALSE || this->async.maxQueDepth < 1 )
		return -1;
	this->AsyncWait();
	for ( i = 0; i < this->async.maxQueDepth; i++ )
	{
		if ( this->async.iotbl[i].hEvent != NULL )
		{
			::CloseHandle(this->async.iotbl[i].hEvent);
			this->async.iotbl[i].hEvent = NULL;
		}
		if ( this->async.iotbl[i].buffer != NULL )
		{
			::VirtualFree( this->async.iotbl[i].buffer, this->ben.maxBufferSize, MEM_DECOMMIT );
			::VirtualFree( this->async.iotbl[i].buffer, this->ben.maxBufferSize, MEM_RELEASE  );
			this->async.iotbl[i].buffer = NULL;
		}
		this->async.iotbl[i].exec = FALSE;
	}
	if ( this->async.iotbl != NULL )
	{
		delete [] this->async.iotbl;
		this->async.iotbl = NULL;
	}
	this->async.queNo = 0;

	return 0;
}
#endif

int		DevTest::fprintf(FILE * fp, const char * form, ... )
{
	int		res;
	char	buffer[TForm::BUFFER_SIZE];
	va_list	arg;

	va_start(arg, form);
	res = TForm::format(sizeof(buffer), buffer, form, arg);
	va_end(arg);
	fputs(buffer, fp);
	return res;
}

char *	DevTest::Fmt(char * buffer, const char * form, ... )
{
	va_list	arg;

	va_start(arg, form);
#if 1
	TForm::format(TForm::BUFFER_SIZE, buffer, form, arg);
#else
	vsprintf(buffer, form, arg);
#endif
	va_end(arg);
	return buffer;
}

//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
// "### [## ## ##]-[## ## ##] ########## ########## ##:#################### #######
// "                                     ########## ########## -------------------
void	DevTest::PutFmt(int spc, const char * text1, const char * val, const char *text2)
{
	fprintf(stdout, "%*s%-*s %10s %s\n"
	  ,spc, ""
	  ,48 - spc, text1
	  ,val
	  ,text2
	);
}

void	DevTest::PutPos(int pos, ... )
{
	va_list	arg;
	char *	text;
	int		x = 0, len;

	va_start(arg, pos);
	while ( pos > 0 )
	{
		text = va_arg(arg, char*);
		len = strlen(text);
		if ( pos > x )
		{
			fprintf(stdout, "%*s", pos - x, "");
			x = pos;
		}
		fprintf(stdout, "%s", text);
		x += len;
		pos  = va_arg(arg, int);
	}
	if ( pos == 0 )
		fprintf(stdout,"\n");
	va_end(arg);
}

int		DevTest::GetNonZeroSize(unsigned size, BYTE * buffer) // static
{
	unsigned	i;

	for ( i = 0; i < size; i++ )
	{
		if ( buffer[size-1-i] != 0 )
			break;
	}
	return size - i;
}

void		DevTest::PrintDumpOct(const char * msg, int ascii, unsigned size, BYTE * buffer)
{
	unsigned	n;

	if ( size == 0 )
		return;
	//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
	// "    #################################  ####: ## ## ## ## ## ## ## ##  ########
	fprintf(stdout,"    %-33s  %04X:", msg, 0 );
	for ( n = 0; n < ((size + 8)/8)*8; n++ )
	{
		if ( n > 0 && (n % 8) == 0 )
			fprintf(stdout,"    %-33s  %04X:", "", n );
		if ( n < size )
			fprintf(stdout, " %02X", buffer[n]);
		else
			fprintf(stdout, "   ");
		if ( (n % 8) == 7 )
		{
			if ( ascii )
			{
				unsigned	i, i8;

				fprintf(stdout, "  ");
				i8 = (n/8)*8;
				for ( i = i8; i < size && i < i8+8; i++ )
					fprintf(stdout, "%c", isprint(buffer[i]) ? buffer[i] : '.');
			}
			fprintf(stdout, "\n");
			if ( n + 1 >= size )
				break;
		}
	}
}

void		DevTest::PrintDumpHex(unsigned addr, unsigned blockSize, unsigned size, BYTE * buffer)
{
	unsigned	x, y;

	if ( size == 0 )
		return;
	//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
	// "ADDRESS    +0 +1 +2 +3 +4 +5 +6 +7 - +8 +9 +A +B +C +D +E +F  ----------------
	// "XXXX-XXXX  ## ## ## ## ## ## ## ## - ## ## ## ## ## ## ## ##  ################
	fprintf(stdout,"ADDRESS    +0 +1 +2 +3 +4 +5 +6 +7 - +8 +9 +A +B +C +D +E +F  ----------------\n");
	for ( y = 0; y < size; y += 16 )
	{
		fprintf(stdout,"%04X-%04X ", (addr+y)>>16, (addr+y) & 0xFFFF);
		for ( x = 0; x < 16; x++ )
		{
			unsigned	c;

			if ( y + x < size )
			{
				c = buffer[y + x];
				fprintf(stdout, " %02X", c);
			} else
				fprintf(stdout, "   ");
			if ( x == 7 )
				fprintf(stdout," -");
		}
		fprintf(stdout,"  ");
		for ( x = 0; x < 16; x++ )
		{
			unsigned	c;

			if ( y + x < size )
			{
				c = buffer[y + x];
				c = isprint(c) ? c : '.';
			} else
				c= ' ';
			fprintf(stdout,"%c", c);
		}
		fprintf(stdout,"\n");
	}
}

#if	defined(DEVTEST_DUMP_CSTYLE)
void		DevTest::PrintDumpCStyle(unsigned size, BYTE * buffer)
{
	unsigned	x, y;

	if ( size == 0 )
		return;
	fprintf(stdout, "{\n");
	for ( y = 0; y < size; y += 8 )
	{
		fprintf(stdout,"\t");
		for ( x = 0; x < 8; x++ )
		{
			unsigned	c;

			if ( y + x < size )
			{
				c = buffer[y + x];
				fprintf(stdout, "0x%02X,", c);
			} else
				fprintf(stdout, "     ");
		}
		fprintf(stdout,"\t// ");	// comment & Ascii
		for ( x = 0; x < 8; x++ )
		{
			unsigned	c;

			if ( y + x < size )
			{
				c = buffer[y + x];
				c = isprint(c) ? c : '.';
			} else
				c= ' ';
			fprintf(stdout,"%c", c);
		}
		fprintf(stdout,"\n");
	}
	fprintf(stdout, "};\n");
}
#endif

char *	DevTest::FormatSize(char * buffer, INT64 size)
{
	unsigned	n, m;
	INT64		ub;

	switch ( this->opt.byteUnit )
	{
		case BYTEUNIT_210:
			ub = I64(1024);
			break;
		case BYTEUNIT_103:
		default:
			ub = I64(1000);
			break;
	}

	if ( size < ub )
	{
		sprintf(buffer, "%4u.0B", (unsigned)size);
	} else if ( size < ub * ub )
	{	// Kilo
		n = (unsigned)(size / ub);
		m = (unsigned)(I64(10) * (size % ub) / ub);
		sprintf(buffer, "%4u.%1u%c", n, m, (this->opt.byteUnit == BYTEUNIT_210 ? 'K' : 'k'));
	} else if ( size < ub * ub * ub )
	{	// Mega
		n = (unsigned)(size / (ub*ub));
		m = (unsigned)(I64(10) * (size % (ub*ub)) / (ub*ub));
		sprintf(buffer, "%4u.%1uM", n, m);
	} else if ( size < ub * ub * ub * ub )
	{	// Giga
		n = (unsigned)(size / (ub*ub*ub));
		m = (unsigned)(I64(10) * (size % (ub*ub*ub)) / (ub*ub*ub));
		sprintf(buffer, "%4u.%1uG", n, m);
	} else if ( size < ub * ub * ub * ub * ub )
	{	// Tela
		n = (unsigned)(size / (ub*ub*ub*ub));
		m = (unsigned)(I64(10) * (size % (ub*ub*ub*ub)) / (ub*ub*ub*ub));
		sprintf(buffer, "%4u.%1uT", n, m);
	}

	return buffer;
}

char *	DevTest::FormatTimeRemain(char * buffer, unsigned sec)
{
	if ( sec < 60 )
		sprintf(buffer, "         %02us", sec);
	else if ( sec < 5 * 60 )
		sprintf(buffer, "     %02um %02us", sec/60, sec%60);
	else if ( sec < 60 * 60 )
		sprintf(buffer, "     %02um    ", sec/60 );
	else
		sprintf(buffer, "%3uh %02um    ", sec/60/60, (sec/60)%60);
	return buffer;
}

char *	DevTest::FormatTime(char * buffer, unsigned sec)	// for devtest_udf.lib
{
#if 0
	if ( sec < 60 )
		sprintf(buffer, "%02us", sec);
	else if ( sec < 60 * 60)
		sprintf(buffer, "%2um %02us", sec/60, sec%60);
	else if ( sec < 24 * 60 * 60 )
		sprintf(buffer, "%2uh %02um %02us", sec/60/60, (sec/60)%60, sec%60 );
	else
		sprintf(buffer, "%3ud %02uh %02um %02us", sec/60/60/24, (sec/60/60)%24, (sec/60)%60, sec%60);
#else
	sprintf(buffer, "%d %02u:%02u:%02u", sec/60/60/24, (sec/60/60)%24, (sec/60)%60, sec%60);
#endif
	return buffer;
}

INT64	DevTest::GetNumberSize(const char *s, unsigned umode)
{
	INT64	n = 0;
	int		sign = 1;	

	if ( *s == '-' )
	{
		sign = -1;
		s++;
	}
	while ( *s )
	{
		if ( *s < '0' || *s > '9' )
			break;
		n = n * I64(10) + (INT64)(*s - '0');
		s++;
	}

	switch ( *s )
	{
		case 'k':
		case 'K':
			switch ( umode )
			{
				case DevTest::BYTEUNIT_103: n = n * I64(1000); break;
				case DevTest::BYTEUNIT_210:
				default                   : n = n * I64(0x400); break;
			}
			break;
		case 'M':
		case 'm':
			switch ( umode )
			{
				case DevTest::BYTEUNIT_103: n = n * I64(1000000); break;
				case DevTest::BYTEUNIT_210:
				default                   : n = n * I64(0x100000); break;
			}
			break;
		case 'G':
		case 'g':
			switch ( umode )
			{
				case DevTest::BYTEUNIT_103: n = n * I64(1000000000); break;
				case DevTest::BYTEUNIT_210:
				default                   : n = n * I64(0x40000000); break;
			}
			break;
	}
	return sign < 0 ? -n : n;
}

#if	defined(WIN32)
char *	DevTest::GetErrMsg(int errcode)
{
	char	*msgbuf;
	int		ret;

	ret = ::FormatMessage(
	    FORMAT_MESSAGE_ALLOCATE_BUFFER |
	    FORMAT_MESSAGE_FROM_SYSTEM |
	    FORMAT_MESSAGE_IGNORE_INSERTS,
	    NULL,
	    errcode,
	    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
	    (LPTSTR) &msgbuf,
	    0,
	    NULL
		);
	return ret > 0 ? msgbuf : "";
}
#endif

char *	DevTest::GetPartitionType(unsigned type)
{
	static const char *tbl[] = 
	{
	/* 0x00 */	"Empty",
	/* 0x01 */	"FAT12",
	/* 0x02 */	"XENIX root",
	/* 0x03 */	"XENIX user",
	/* 0x04 */	"FAT16 (<32M)",
	/* 0x05 */	"Extended",
	/* 0x06 */	"FAT16 (>=32M)",
	/* 0x07 */	"NTFS",
	/* 0x08 */	"OS/2",
	/* 0x09 */	"AIX data",
	/* 0x0A */	"OS/2 Boot Manager",
	/* 0x0B */	"FAT32",
	/* 0x0C */	"FAT32 LBA",
	/* 0x0D */	"",
	/* 0x0E */	"FAT16 LBA",
	/* 0x0F */	"Extended LBA",

	/* 0x10 */	"",
	/* 0x11 */	"Hidden FAT12",
	/* 0x12 */	"",
	/* 0x13 */	"",
	/* 0x14 */	"Hidden FAT16 (<32M)",
	/* 0x15 */	"",
	/* 0x16 */	"Hidden FAT16 (>=32M)",
	/* 0x17 */	"Hidden HPFS",
	/* 0x18 */	"",
	/* 0x19 */	"",
	/* 0x1A */	"",
	/* 0x1B */	"Hidden FAT32 Hidden",
	/* 0x1C */	"Hidden FAT32 LBA",
	/* 0x1D */	"",
	/* 0x1E */	"Hidden FAT16 LBA",
	/* 0x1F */	"",

	/* 0x20 */	"",
	/* 0x21 */	"",
	/* 0x22 */	"",
	/* 0x23 */	"",
	/* 0x24 */	"NEC DOS 3.x",
	/* 0x25 */	"",
	/* 0x26 */	"",
	/* 0x27 */	"",
	/* 0x28 */	"",
	/* 0x29 */	"",
	/* 0x2A */	"",
	/* 0x2B */	"",
	/* 0x2C */	"",
	/* 0x2D */	"",
	/* 0x2E */	"",
	/* 0x2F */	"",

	/* 0x30 */	"",
	/* 0x31 */	"",
	/* 0x32 */	"",
	/* 0x33 */	"",
	/* 0x34 */	"",
	/* 0x35 */	"",
	/* 0x36 */	"",
	/* 0x37 */	"",
	/* 0x38 */	"",
	/* 0x39 */	"Paln9",
	/* 0x3A */	"",
	/* 0x3B */	"",
	/* 0x3C */	"PartitionMagic",
	/* 0x3D */	"Hidden NetWare",
	/* 0x3E */	"",
	/* 0x3F */	"",

	/* 0x40 */	"Venix 80286",
	/* 0x41 */	"Linux/MINIX",
	/* 0x42 */	"Linux swap",
	/* 0x43 */	"Linux native",
	/* 0x44 */	"GoBack",
	/* 0x45 */	"",
	/* 0x46 */	"",
	/* 0x47 */	"",
	/* 0x48 */	"",
	/* 0x49 */	"",
	/* 0x4A */	"",
	/* 0x4B */	"",
	/* 0x4C */	"",
	/* 0x4D */	"",
	/* 0x4E */	"",
	/* 0x4F */	"",

	/* 0x50 */	"",
	/* 0x51 */	"",
	/* 0x52 */	"CP/M",
	/* 0x53 */	"",
	/* 0x54 */	"",
	/* 0x55 */	"EZ-Drive",
	/* 0x56 */	"",
	/* 0x57 */	"",
	/* 0x58 */	"",
	/* 0x59 */	"",
	/* 0x5A */	"",
	/* 0x5B */	"",
	/* 0x5C */	"",
	/* 0x5D */	"",
	/* 0x5E */	"",
	/* 0x5F */	"",

	/* 0x60 */	"",
	/* 0x61 */	"",
	/* 0x62 */	"",
	/* 0x63 */	"",
	/* 0x64 */	"Novell NetWare 286",
	/* 0x65 */	"Novell NetWare 386",
	/* 0x66 */	"Novell NetWare SMS",
	/* 0x67 */	"Novell",
	/* 0x68 */	"Novell",
	/* 0x69 */	"Novell",
	/* 0x6A */	"",
	/* 0x6B */	"",
	/* 0x6C */	"",
	/* 0x6D */	"",
	/* 0x6E */	"",
	/* 0x6F */	"",

	/* 0x70 */	"",
	/* 0x71 */	"",
	/* 0x72 */	"",
	/* 0x73 */	"",
	/* 0x74 */	"",
	/* 0x75 */	"IBM PC/IX",
	/* 0x76 */	"",
	/* 0x77 */	"",
	/* 0x78 */	"",
	/* 0x79 */	"",
	/* 0x7A */	"",
	/* 0x7B */	"",
	/* 0x7C */	"",
	/* 0x7D */	"",
	/* 0x7E */	"",
	/* 0x7F */	"",

	/* 0x80 */	"MINIX",
	/* 0x81 */	"MINUX",
	/* 0x82 */	"Linux swap/Solaris86",
	/* 0x83 */	"Linix native",
	/* 0x84 */	"Hidden partition",
	/* 0x85 */	"Linux extended",
	/* 0x86 */	"NTFS volume set",
	/* 0x87 */	"",
	/* 0x88 */	"",
	/* 0x89 */	"",
	/* 0x8A */	"",
	/* 0x8B */	"",
	/* 0x8C */	"",
	/* 0x8D */	"",
	/* 0x8E */	"",
	/* 0x8F */	"",

	/* 0x90 */	"",
	/* 0x91 */	"",
	/* 0x92 */	"",
	/* 0x93 */	"",
	/* 0x94 */	"",
	/* 0x95 */	"",
	/* 0x96 */	"",
	/* 0x97 */	"",
	/* 0x98 */	"",
	/* 0x99 */	"",
	/* 0x9A */	"",
	/* 0x9B */	"",
	/* 0x9C */	"",
	/* 0x9D */	"",
	/* 0x9E */	"",
	/* 0x9F */	"BSD/OS",

	/* 0xA0 */	"",
	/* 0xA1 */	"",
	/* 0xA2 */	"",
	/* 0xA3 */	"",
	/* 0xA4 */	"",
	/* 0xA5 */	"BSD/386",
	/* 0xA6 */	"OpenBSD",
	/* 0xA7 */	"NEXTSTEP",
	/* 0xA8 */	"Mac OS-X",
	/* 0xA9 */	"NetBSD",
	/* 0xAA */	"",
	/* 0xAB */	"Mac OS-X boot",
	/* 0xAC */	"",
	/* 0xAD */	"",
	/* 0xAE */	"",
	/* 0xAF */	"",

	/* 0xB0 */	"",
	/* 0xB1 */	"",
	/* 0xB2 */	"",
	/* 0xB3 */	"",
	/* 0xB4 */	"",
	/* 0xB5 */	"",
	/* 0xB6 */	"NT mirror FAT16",
	/* 0xB7 */	"NT mirror NTFS",
	/* 0xB8 */	"",
	/* 0xB9 */	"",
	/* 0xBA */	"",
	/* 0xBB */	"",
	/* 0xBC */	"",
	/* 0xBD */	"",
	/* 0xBE */	"",
	/* 0xBF */	"",

	/* 0xC0 */	"",
	/* 0xC1 */	"",
	/* 0xC2 */	"Hidden Linux",
	/* 0xC3 */	"Hidden Linux swap",
	/* 0xC4 */	"",
	/* 0xC5 */	"",
	/* 0xC6 */	"NT FAT vol/stripe",
	/* 0xC7 */	"NT NTFS vol/stripe",
	/* 0xC8 */	"",
	/* 0xC9 */	"",
	/* 0xCA */	"",
	/* 0xCB */	"",
	/* 0xCC */	"",
	/* 0xCD */	"",
	/* 0xCE */	"",
	/* 0xCF */	"",

	/* 0xD0 */	"",
	/* 0xD1 */	"",
	/* 0xD2 */	"",
	/* 0xD3 */	"",
	/* 0xD4 */	"",
	/* 0xD5 */	"",
	/* 0xD6 */	"",
	/* 0xD7 */	"",
	/* 0xD8 */	"",
	/* 0xD9 */	"",
	/* 0xDA */	"",
	/* 0xDB */	"",
	/* 0xDC */	"",
	/* 0xDD */	"",
	/* 0xDE */	"",
	/* 0xDF */	"",

	/* 0xE0 */	"",
	/* 0xE1 */	"",
	/* 0xE2 */	"",
	/* 0xE3 */	"",
	/* 0xE4 */	"",
	/* 0xE5 */	"",
	/* 0xE6 */	"",
	/* 0xE7 */	"",
	/* 0xE8 */	"",
	/* 0xE9 */	"",
	/* 0xEA */	"",
	/* 0xEB */	"",
	/* 0xEC */	"",
	/* 0xED */	"BeOS",
	/* 0xEE */	"",
	/* 0xEF */	"",

	/* 0xF0 */	"",
	/* 0xF1 */	"",
	/* 0xF2 */	"",
	/* 0xF3 */	"",
	/* 0xF4 */	"",
	/* 0xF5 */	"",
	/* 0xF6 */	"",
	/* 0xF7 */	"",
	/* 0xF8 */	"",
	/* 0xF9 */	"",
	/* 0xFA */	"",
	/* 0xFB */	"VMware File System",
	/* 0xFC */	"VMware Swap",
	/* 0xFD */	"",
	/* 0xFE */	"NT Disk Admin hidden",
	/* 0xFF */	"",
	};
	return (char *)tbl[type & 0xFF];
}

DevTest::Clock_t	DevTest::Clock(void)
{
#if	defined(WIN32)
	return (Clock_t)::clock();
#endif
#if	defined(LINUX)
	struct timeval	tv;

	gettimeofday(&tv, NULL);
	return (Clock_t)tv.tv_sec * CLOCKS_PER_SEC + (Clock_t)tv.tv_usec * CLOCKS_PER_SEC / I64(1000000);
#endif
}

#if	defined(WIN32)
int		DevTest::InitCpuUsage(void)
{
	PDH_COUNTER_PATH_ELEMENTS	ele;

	memset(&ele, 0, sizeof(ele));
	switch ( this->opt.cpuUsage )
	{
		case CPUUSAGE_ALL:
#if	defined(DEBUG)
			fprintf(stderr, "CPU USAGE: ALL\n");	// DEBUG
#endif
			ele.szObjectName   = TEXT("Processor");
			ele.szInstanceName = TEXT("_Total");
			ele.szCounterName  = TEXT("% Processor Time");
			break;
		case CPUUSAGE_SELF:
		{
			char	path[MAX_PATH], drive[MAX_PATH], dir[MAX_PATH], fname[MAX_PATH], ext[MAX_PATH];

			::GetModuleFileName(NULL, path, sizeof(path));
			::_splitpath(path, drive, dir, fname, ext);
#if	defined(DEBUG)
			fprintf(stderr, "CPU USAGE: SELF (%s)\n", fname);	// DEBUG
#endif
			ele.szObjectName   = TEXT("Process");
			ele.szInstanceName = fname;
			ele.szCounterName  = TEXT("% Processor Time");
			break;
		}
		default:
			return -1;
	}
	this->perflib.OpenQuery();
	if ( (this->counter = this->perflib.AddCounter(ele, NULL)) == NULL )
	{
		fprintf(stderr, "PerfLib::AddCounter error!: %s\n", this->perflib.GetErrMsg() );
		return -1;	// error
	}
	return 0;
}

void	DevTest::CollectCpuUsage(void)
{
	this->perflib.CollectQueryData();
}

double	DevTest::GetCpuUsage(void)
{
	PDH_FMT_COUNTERVALUE	value;

	this->perflib.GetFormattedCounterValue(this->counter, PDH_FMT_DOUBLE, NULL, &value);
	return value.doubleValue;
}

void	DevTest::TermCpuUsage(void)
{
	this->perflib.CloseQuery();
}

#elif defined(LINUX)
//////////////////////////////////////////////////////////////////////////
//	CPU USAGE for Linux

static DevTest::Clock_t	tclk0, tclk1, tclk, rclk;
static struct rusage	rus0, rus1;

#define	TIMEVAL_TO_CLOCK(tm)	(Clock_t)tm.tv_sec * CLOCKS_PER_SEC + (Clock_t)tm.tv_usec * CLOCKS_PER_SEC / I64(1000000)

int		DevTest::InitCpuUsage(void)
{
	tclk0 = this->Clock();
	getrusage(RUSAGE_SELF, &rus0);
	return 0;
}

void	DevTest::CollectCpuUsage(void)
{
	getrusage(RUSAGE_SELF, &rus1);
	tclk1 = this->Clock();

	tclk = tclk1 - tclk0;
	switch ( this->opt.cpuUsage )
	{
		case CPUUSAGE_SYS :
			rclk = TIMEVAL_TO_CLOCK(rus1.ru_stime) - TIMEVAL_TO_CLOCK(rus0.ru_stime);
			break;
		case CPUUSAGE_USER:
			rclk = TIMEVAL_TO_CLOCK(rus1.ru_utime) - TIMEVAL_TO_CLOCK(rus0.ru_utime);
			break;
		default:
			rclk  = (TIMEVAL_TO_CLOCK(rus1.ru_utime) + TIMEVAL_TO_CLOCK(rus1.ru_stime))
			      - (TIMEVAL_TO_CLOCK(rus0.ru_utime) + TIMEVAL_TO_CLOCK(rus0.ru_stime));
			break;
	}
	tclk0 = tclk1;
	rus0.ru_utime = rus1.ru_utime;
	rus0.ru_stime = rus1.ru_stime;
}

double	DevTest::GetCpuUsage(void)
{
	if ( tclk == 0 )
		return 0.0;
	return 100.0 * (double)rclk / (double)tclk;
}

void	DevTest::TermCpuUsage(void)
{
}
#endif
