/////////////////////////////////////////////////////////////////////////f
//	"devtest_smart.cpp"
//------------------------------------------------------------------------
//	$Id: devtest_smart.cpp 164 2006-03-21 10:32:03Z taro $
//////////////////////////////////////////////////////////////////////////

#include	<stdlib.h>
#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	<ctype.h>
#include	"devtest.h"
#include	"defs.h"
#include	"tform.h"

#if	defined(WIN32)
# include	<crtdbg.h>
#elif defined(LINUX)
# include	<assert.h>
# define	_ASSERT
#endif

#if defined(_DEBUG) && defined(WIN32)
# define	DBGREPORT(param_)	_CrtDbgReport param_
#else
# define	DBGREPORT(param_)
#endif

int		DevTest::PrintSmartInfo(DevTest::DevParam & devParam)
{
#if	defined(WIN32)
	GETVERSIONINPARAMS			ver;
#endif
	DevCtl::AtaIdentifyDevice	identify;
	DevCtl::AtaSmartValues		smart_val;
	DevCtl::AtaSmartThresholds	smart_thr;
	SmartVendor					vendor = SMART_VENDOR_GENERAL;
	TCHAR						t1[80], t2[80];
	int							drive = 0;
	int							val_ok = FALSE, thr_ok = FALSE;

#if	defined(WIN32)
	if ( this->devCtl.SmartGetVersion(ver) == 0 )
	{
		static struct
		{
			unsigned		bit;
			const TCHAR *	text;
		} idemap[] =
		{
			{ 0x01, TEXT("IDE Primary master"    ) },
			{ 0x02, TEXT("IDE Primary slave"     ) },
			{ 0x04, TEXT("IDE Secondary master"  ) },
			{ 0x08, TEXT("IDE Secondary slave"   ) },
			{ 0x10, TEXT("ATAPI Primary master"  ) },
			{ 0x20, TEXT("ATAPI Primary slave"   ) },
			{ 0x40, TEXT("ATAPI Secondary master") },
			{ 0x80, TEXT("ATAPI Secondary slave" ) },
		};
		int				i;

		PutPos(4        , TEXT("S.M.A.R.T. Version"),
		       COLUMN3-21, Fmt(t1,"%20u", (unsigned)ver.bVersion),
		       0);
		PutPos(8        , TEXT("IDE Device map"),
		       COLUMN3-3, Fmt(t1,"%02XH", (unsigned)ver.bIDEDeviceMap),
		       0);
		for ( i = 0; i < sizeof(idemap)/sizeof(*idemap); i++ )
		{
			if ( ver.bIDEDeviceMap & idemap[i].bit )
			{
				PutPos(COLUMN3 - _tcslen(idemap[i].text) - 1, idemap[i].text, 0);
			}
		}
	}
#endif

	if ( this->devCtl.AtaReadIndentify(drive, identify) == 0 )
	{
		TCHAR		model[40+1], firm[8+1], serial[20+1];
		INT64		maxlba;
		unsigned	buffer_size;;

		vendor = DevTest::GetAtaVendor(identify);

		if ( this->opt.printDetail )
			this->PrintDumpHex(0, 512, sizeof(identify), (BYTE *)&identify);

		FormatAtaString(model, identify.model_number, sizeof(identify.model_number));
		FormatAtaString(firm, identify.micro_code_version, sizeof(identify.micro_code_version));
		FormatAtaString(serial, identify.serial_number, sizeof(identify.serial_number));
		maxlba = DevTest::GetAtaCapacity(identify);
		buffer_size = DevTest::ATA_GET16(&identify.buffer_size);

		{
			DispParam tbl[] =
			{
				{ DISPTYPE_SHORT_TEXT    , TEXT("Model"         ), (void *)model       , 0, NULL},
				{ DISPTYPE_SHORT_TEXT    , TEXT("Firmware"      ), (void *)firm        , 0, NULL},
				{ DISPTYPE_SHORT_TEXT    , TEXT("Serial Number" ), (void *)serial      , 0, NULL},
				{ DISPTYPE_LBA64_CAPACITY, TEXT("Capacity (LBA)"), (void *)&maxlba     , 0, NULL},
				{ DISPTYPE_UINT32_512SIZE, TEXT("Buffer size"   ), (void *)&buffer_size, 0, NULL},
			};
			this->DisplayInfo(sizeof(tbl)/sizeof(*tbl), tbl);
		}
		if ( this->opt.printDetail )
		{
			unsigned	stt = identify.device_classificaion.GetData();
			const TCHAR	*s;

			PutPos(8, TEXT("Device classificaion")
			      ,COLUMN3 - 5 -1, Fmt(t1, TEXT("%04XH"), stt )
			      ,0);
			if ( stt & (1<<15) )
				s = TEXT("ATAPI drive");
			else
				s = TEXT("ATA drive");
			PutPos(COLUMN3 - _tcslen(s) - 1, s
			      ,0);

			if ( stt & (1<<7) )
			{
				s = TEXT("Removable cartridge drive");
				PutPos(COLUMN3 - _tcslen(s) - 1, s
				      ,0);
			}
			if ( stt & (1<<6) )
			{
				s = TEXT("Fixed drive");
				PutPos(COLUMN3 - _tcslen(s) - 1, s
				      ,0);
			}
			if ( stt & (1<<2) )
			{
				s = TEXT("Soft sectored");
				PutPos(COLUMN3 - _tcslen(s) - 1, s
				      ,0);
			}
		}
		if ( this->opt.printDetail )
		{
			static DispParamBitDef tbl82[] =
			{
				{ 14, 1, TEXT("NOP command") },
				{ 13, 1, TEXT("READ BUFFER command") },
				{ 12, 1, TEXT("WRITE BUFFER command") },
				{ 10, 1, TEXT("Host Protected feature set") },
				{  9, 0, TEXT("DEVICE RESET command") },
				{  8, 0, TEXT("SERVICE interrupt") },
				{  7, 1, TEXT("RELEASE interrupt") },
				{  6, 1, TEXT("LOOK AHEAD") },
				{  5, 1, TEXT("WRITE CACHE") },
				{  4, 0, TEXT("PACKAET command fearure set") },
				{  3, 1, TEXT("Power Management fearure set") },
				{  2, 0, TEXT("Removable fearure set") },
				{  1, 1, TEXT("Security fearure set") },
				{  0, 1, TEXT("SMART fearure set") },
			};
			static DispParamBitDef tbl83[] =
			{
				{ 13, 1, TEXT("FLUSH CACHE EXT command") },
				{ 12, 1, TEXT("FLUSH CACHE command") },
				{ 11, 1, TEXT("Device Configuration Overlay command") },
				{ 10, 1, TEXT("48-bit Address feature set") },
				{  9, 1, TEXT("Automatic Acoustic mode") },
				{  8, 1, TEXT("SET Max Security extension") },
				{  7, 1, TEXT("Set Features Address Offset feature mode") },
				{  6, 1, TEXT("SET FEATURES subcommand required to spin-up") },
				{  5, 1, TEXT("Power-Up In Standby feature set supported") },
				{  4, 0, TEXT("Removable Media Status Notification feature") },
				{  3, 1, TEXT("Advanced Power Management feature set") },
				{  2, 1, TEXT("CFA feature set") },
				{  1, 1, TEXT("READ/WRITE DMA QUEUED") },
				{  0, 0, TEXT("DOWNLOAD MICROCODE command") },
			};
			static DispParamBitDef tbl84[] =
			{
				{  5, 1, TEXT("General purpose logging feature set") },
				{  1, 1, TEXT("SMART self-test supported") },
				{  0, 0, TEXT("SMART error logging supported") },
			};
			static DispParamBitDef tbl85[] =
			{
				{ 14, 1, TEXT("NOP command") },
				{ 13, 1, TEXT("READ BUFFER command") },
				{ 12, 1, TEXT("WRITE BUFFER command") },
				{ 10, 1, TEXT("Host Protected Area Feature set") },
				{  9, 1, TEXT("DEVICE RESET command") },
				{  8, 1, TEXT("SERVICE interrupt") },
				{  7, 1, TEXT("RELEASE interrupt") },
				{  6, 1, TEXT("LOOK AHEAD") },
				{  5, 1, TEXT("WRITE CACHE") },
				{  4, 1, TEXT("PACKAET command fearure set") },
				{  3, 1, TEXT("Power Management fearure set") },
				{  2, 1, TEXT("Removable fearure set") },
				{  1, 1, TEXT("Security fearure set") },
				{  0, 1, TEXT("SMART fearure set") },
			};
			static DispParamBitDef tbl86[] =
			{
				{ 13, 1, TEXT("FLUSH CACHE EXT command") },
				{ 12, 1, TEXT("FLUSH CACHE command") },
				{ 11, 1, TEXT("Device Configuration Overlay command") },
				{ 10, 1, TEXT("48-bit Address feature set") },
				{  9, 1, TEXT("Automatic Acoustic mode") },
				{  8, 1, TEXT("SET Max Security extension") },
				{  7, 1, TEXT("Set Features Address Offset feature mode") },
				{  6, 1, TEXT("SET FEATURES subcommand required to spin-up after power-up") },
				{  5, 1, TEXT("Power-Up In Standby feature set supported") },
				{  4, 1, TEXT("Removable Media Status Notification feature") },
				{  3, 1, TEXT("Advanced Power Management feature set") },
				{  2, 1, TEXT("CFA feature set") },
				{  1, 1, TEXT("READ/WRITE DMA QUEUED") },
				{  0, 1, TEXT("DOWNLOAD MICROCODE command") },
			};
			static DispParamBitDef tbl87[] =
			{
				{  5, 1, TEXT("General purpose logging feature set") },
				{  1, 1, TEXT("SMART self-test supported") },
				{  0, 1, TEXT("SMART error logging supported") },
			};
			WORD		command82 = identify.command_set082.GetData();
			WORD		command83 = identify.command_set083.GetData();
			WORD		command84 = identify.command_set_extension.GetData();
			WORD		command85 = identify.command_set_enable.GetData();
			WORD		command86 = identify.command_set086.GetData();
			WORD		command87 = identify.command_set087.GetData();

			DispParam	disptbl[] =
			{
				{ DISPTYPE_UINT16_BITDEF, TEXT("Command set supported"      ), (void *)&command82, sizeof(tbl82)/sizeof(*tbl82), tbl82},
				{ DISPTYPE_UINT16_BITDEF, TEXT("Command set supported"      ), (void *)&command83, sizeof(tbl83)/sizeof(*tbl83), tbl83},
				{ DISPTYPE_UINT16_BITDEF, TEXT("Command set extension"      ), (void *)&command84, sizeof(tbl84)/sizeof(*tbl84), tbl84},
				{ DISPTYPE_UINT16_BITDEF, TEXT("Command set/feature enabled"), (void *)&command85, sizeof(tbl85)/sizeof(*tbl85), tbl85},
				{ DISPTYPE_UINT16_BITDEF, TEXT("Command set/feature enabled"), (void *)&command86, sizeof(tbl86)/sizeof(*tbl86), tbl86},
				{ DISPTYPE_UINT16_BITDEF, TEXT("Command set/feature enabled"), (void *)&command87, sizeof(tbl87)/sizeof(*tbl87), tbl87},
			};
			this->DisplayInfo(sizeof(disptbl)/sizeof(*disptbl), disptbl);
		}
		if ( this->opt.printDetail )
		{
			unsigned	udma_mode = identify.ultra_DMA_trasnfer_mode.GetData();
			int			n;
			DispParam	dispparam = { DISPTYPE_UINT16_BITS, TEXT("Ultra DMA transfer mode"), (void *)&udma_mode, 0, NULL};
			static const TCHAR	*modemsg_tbl[] =
			{
				TEXT(""),
				TEXT(""),
				TEXT(" 33"),
				TEXT(""),
				TEXT(" 66"),
				TEXT(" 100"),
				TEXT(" 133"),
				TEXT(""),
			};

			this->DisplayInfo(1, &dispparam);
			for ( n = 7; n >= 0; n-- )
			{
				if ( udma_mode & (1<<(8+n)) )
				{
					Fmt(t1, TEXT("Mode %d"), n);
					PutPos(12, TEXT("Current")
					      ,COLUMN3 - _tcslen(t1)-1, t1
					      ,COLUMN3, Fmt(t2, TEXT("Active"))
					      ,0);
				}
			}
			for ( n = 7; n >= 0; n-- )
			{
				if ( udma_mode & (1<<n) )
				{
					Fmt(t1, TEXT("Mode %d"),  n);
					PutPos(COLUMN3 - _tcslen(t1)-1, t1
					      ,COLUMN3, Fmt(t2, TEXT("Support%s"), modemsg_tbl[n])
					      ,0);
				}
			}
		}

	}

	if ( this->devCtl.SmartReadValues(drive, smart_val) == 0 )
	{
		val_ok = TRUE;
	//	this->PrintDumpHex(0, 512, sizeof(smart_val), (BYTE *)&smart_val);
	}

	if ( this->devCtl.SmartReadThresholds(drive, smart_thr) == 0 )
	{
		thr_ok = TRUE;
	//	this->PrintDumpHex(0, 512, sizeof(smart_thr), (BYTE *)&smart_thr);
	}

	//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
	// "    ID  Attributes                     Val Wst Ths Flags RAW"
	// "    ### ############################## ### ### ### ####H RAW"
	if ( val_ok )
	{
		int				i;

		PutPos(4     , TEXT("ID# S.M.A.R.T. Attribute"),
		       4 + 4+SMART_ATTRIBUTE_MAX_LENGHTH, TEXT("Val Wst Ths Flags"),
		       COLUMN_MAX - 1 - 9               , TEXT("RAW Value"),
		       0);

		for ( i = 0; i < DevCtl::NUMBER_ATA_SMART_ATTRIBUTES; i++ )
		{
			DevCtl::AtaSmartAttribute			*cur;
			DevCtl::AtaSmartThresholdAttribute	*thr;
			TCHAR								att_name[80], att_msg[80], raw_msg[80], stt_msg[80];
			SmartAttrItem						*attr_item;
			SmartAttrRawType					rawtype = SMART_RAW_NORMAL;

			cur = NULL;
			thr = NULL;
			if ( smart_val.device_attribute[i].id == 0 )
				continue;	// not use
			cur = &smart_val.device_attribute[i];
			attr_item = this->GetSmartAttrItem(cur->id);
			if ( thr_ok )
			{
				int			j;

				// search threshhold
				for ( j = 0; j < DevCtl::NUMBER_ATA_SMART_ATTRIBUTES; j++ )
				{
					if ( smart_thr.threshhold_attribute[i].id == cur->id )
					{
						thr = &smart_thr.threshhold_attribute[i];
						break;
					}
				}
			}
			if ( attr_item != NULL && attr_item->desc != NULL && _tcslen(attr_item->desc) > 0 )
			{
				_tcsncpy(att_name, attr_item->desc, sizeof(att_name)/sizeof(*att_name) - 1);
			} else
			{
				_tcsncpy(att_name, DevTest::GetSmartAttributeIdName(vendor, cur->id), sizeof(att_name)/sizeof(*att_name) - 1);
			}

			{
				unsigned	stt;

				stt = ATA_GET16(cur->flags);
				Fmt(stt_msg, TEXT("%c%c%c%c%c%c")
				  , stt & 0x0020 ? TEXT('1') : TEXT('-')
				  , stt & 0x0010 ? TEXT('1') : TEXT('-')
				  , stt & 0x0008 ? TEXT('1') : TEXT('-')
				  , stt & 0x0004 ? TEXT('1') : TEXT('-')
				  , stt & 0x0002 ? TEXT('O') : TEXT('-')
				  , stt & 0x0001 ? TEXT('P') : TEXT('-')
				);
			}

			Fmt(att_msg, TEXT("%3u %3u %3s %s")
			           , cur->value
			           , cur->worst
			           , thr != NULL ? Fmt(t1,TEXT("%3u"),thr->threshold) : TEXT("???")
			           , stt_msg
			);

			this->GetSmartRawValue(*cur, sizeof(raw_msg)/sizeof(*raw_msg), raw_msg);

			PutPos(4     , Fmt(t1, TEXT("%3u %s"), cur->id, att_name),
			       4 + 4+SMART_ATTRIBUTE_MAX_LENGHTH, att_msg,
			       COLUMN_MAX - 1 - _tcslen(raw_msg), raw_msg,
			       0);
		}
	}

	return 0;
}

TCHAR * DevTest::FormatAtaString(TCHAR * out, const DevCtl::ATA_WORD * inp, int n) // static
{
	TCHAR	tmp[80];
	int		i;

	n = n > sizeof(tmp)/sizeof(*tmp)-1 ? sizeof(tmp)/sizeof(*tmp)-1 : n;
	for ( i = 0; i < n; i += 2 )
	{
		tmp[i  ] = inp[i/2].b[1];
		tmp[i+1] = inp[i/2].b[0];
	}
	tmp[n] = TEXT('\0');

	if ( _tcslen(tmp) > 0 )
	{
		int			top = 0, end;
		char *		s = tmp;

		for ( i = 0; tmp[i]; i++ )
		{
			if ( tmp[i] != TEXT(' ') && tmp[i] != TEXT('\t') )
			{
				top = i;
				break;
			}
		}
		i = (int)_tcslen(tmp) - 1;
		while ( i >= top )
		{
			if ( tmp[i] != TEXT(' ') && tmp[i] != TEXT('\t') )
				break;
			i--;
		}
		end = i;

		_tcsncpy(out, tmp + top, end - top + 1);
		out[end - top + 1] = TEXT('\0');
	} else
		out[0] = TEXT('\0');
	return out;
}

DevTest::SmartVendor	DevTest::GetAtaVendor(DevCtl::AtaIdentifyDevice & identify)
{
	SmartVendor	vendor = SMART_VENDOR_GENERAL;
	TCHAR		model[40+1];
	enum	CmpType
	{
		CMP_STRCMP ,
		CMP_STRICMP,
	};
	static struct
	{
		CmpType			cmp;
		SmartVendor		vendor;
		const TCHAR		*text;
	} tbl[] =
	{
		{ CMP_STRICMP, SMART_VENDOR_FUJITSU, TEXT("FUJITSU") },
		{ CMP_STRICMP, SMART_VENDOR_IBM    , TEXT("IBM"    ) },
		{ CMP_STRCMP , SMART_VENDOR_IBM    , TEXT("IC3"    ) },
		{ CMP_STRCMP , SMART_VENDOR_HGST   , TEXT("HDS"    ) },
		{ CMP_STRICMP, SMART_VENDOR_MAXTOR , TEXT("MAXTOR" ) },
		{ CMP_STRCMP , SMART_VENDOR_SEAGATE, TEXT("ST3"    ) },
		{ CMP_STRICMP, SMART_VENDOR_TOSHIBA, TEXT("TOSHIBA") },
		{ CMP_STRCMP , SMART_VENDOR_WD     , TEXT("WDC"    ) },
	};
	int				i;

	DevTest::FormatAtaString(model, identify.model_number, sizeof(identify.model_number));
	for ( i = 0; i < sizeof(tbl)/sizeof(*tbl); i++ )
	{
		switch ( tbl[i].cmp )
		{
		case CMP_STRCMP:
			if ( _tcsncmp(model, tbl[i].text, _tcslen(tbl[i].text)) == 0 )
				vendor = tbl[i].vendor;
			break;
		case CMP_STRICMP:
			if ( _tcsnicmp(model, tbl[i].text, _tcslen(tbl[i].text)) == 0 )
				vendor = tbl[i].vendor;
			break;
		}
		if ( vendor != SMART_VENDOR_GENERAL )
			break;
	}
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "Model \"%s\" : Vendor = %d\n", model, vendor));

	return vendor;
}

UINT64	DevTest::GetAtaCapacity(DevCtl::AtaIdentifyDevice & identify) // static
{
	WORD	command_set_2  = identify.command_set083.GetData();
	WORD 	capabilities_0 = identify.capabilities049.GetData();
	WORD	sects_16       = identify.total_user_address_sectors[0].GetData();
	WORD	sects_32       = identify.total_user_address_sectors[1].GetData();
	WORD	lba_16         = identify.min_user_LBA_address_48bit[0].GetData();
	WORD	lba_32         = identify.min_user_LBA_address_48bit[1].GetData();
	WORD	lba_48         = identify.min_user_LBA_address_48bit[2].GetData();
	WORD	lba_64         = identify.min_user_LBA_address_48bit[3].GetData();
	UINT64	capacity = I64(0), capacity_short = I64(0);
  
	if ( capabilities_0 & 0x0200 )
	{
		capacity_short = (DWORD)sects_32 << 16 | 
	                     (DWORD)sects_16 << 0  ;

		if ( (command_set_2 & 0xC000) == 0x4000 && (command_set_2 & 0x0400) )
		{	// support 48bit LBA
			capacity = (UINT64)lba_64 << 48 | 
			           (UINT64)lba_48 << 32 |
			           (UINT64)lba_32 << 16 | 
			           (UINT64)lba_16 << 0  ;
		}

		if ( capacity_short > capacity )
			capacity = capacity_short;
	}

	return capacity;
}


const TCHAR * DevTest::GetSmartAttributeIdName(DevTest::SmartVendor vendor, unsigned id, unsigned length, TCHAR * buffer) // static
{
	// MAX length 30
	static struct
	{
		unsigned	id;
		SmartVendor	vendor;
		const TCHAR	*text;
	} tbl[] =
	{
		{   1, SMART_VENDOR_GENERAL, TEXT("Raw Read Error Rate"           ) },
		{   2, SMART_VENDOR_GENERAL, TEXT("Throughput Performance"        ) },
		{   3, SMART_VENDOR_GENERAL, TEXT("Spin Up Time"                  ) },
		{   4, SMART_VENDOR_GENERAL, TEXT("Start/Stop Count"              ) },
		{   5, SMART_VENDOR_GENERAL, TEXT("Reallocated Sectors Count"     ) },
		{   6, SMART_VENDOR_GENERAL, TEXT("Read Channel Margin"           ) },
		{   7, SMART_VENDOR_GENERAL, TEXT("Seek Error Rate"               ) },
		{   8, SMART_VENDOR_GENERAL, TEXT("Seek Time Performance"         ) },
		{   9, SMART_VENDOR_GENERAL, TEXT("Power-On Hours"                ) },
		{  10, SMART_VENDOR_GENERAL, TEXT("Spin Retry Count"              ) },
		{  11, SMART_VENDOR_GENERAL, TEXT("Recalibration Retries"         ) },
		{  12, SMART_VENDOR_GENERAL, TEXT("Device Power Cycle Count"      ) },
		{  13, SMART_VENDOR_GENERAL, TEXT("Soft Read Error Rate"          ) },
		{ 191, SMART_VENDOR_GENERAL, TEXT("Gsense Error Rate"             ) },
		{ 192, SMART_VENDOR_GENERAL, TEXT("Power off Retract Count"       ) },
	//	{ 192, SMART_VENDOR_FUJITSU, TEXT("Emergency Retract Cycles"      ) },	// Fujitsu
		{ 193, SMART_VENDOR_GENERAL, TEXT("Load/Unload Cycle Count"       ) },
		{ 194, SMART_VENDOR_GENERAL, TEXT("Temperature"                   ) },
		{ 195, SMART_VENDOR_GENERAL, TEXT("Hardware ECC Recovered"        ) },
		{ 196, SMART_VENDOR_GENERAL, TEXT("Reallocation Event Count"      ) },
		{ 197, SMART_VENDOR_GENERAL, TEXT("Current Pending Sector Count"  ) },
		{ 198, SMART_VENDOR_GENERAL, TEXT("Uncorrectable Sector Count"    ) },
		{ 199, SMART_VENDOR_GENERAL, TEXT("UltraDMA CRC Error Rate"       ) },
		{ 200, SMART_VENDOR_WD     , TEXT("Multi Zone Error Rate"         ) },	// Western Digital
		{ 200, SMART_VENDOR_FUJITSU, TEXT("Write Error Rate"              ) },	// Fujitsu
		{ 201, SMART_VENDOR_GENERAL, TEXT("Soft Read Error Rate"          ) },
		{ 202, SMART_VENDOR_MAXTOR , TEXT("Data Address Mark Errors"      ) },	// Maxtor
		{ 202, SMART_VENDOR_FUJITSU, TEXT("TA Increase Count"             ) },	// Fujitsu
		{ 203, SMART_VENDOR_MAXTOR , TEXT("ECC Errors"                    ) },	// Maxtor
		{ 203, SMART_VENDOR_FUJITSU, TEXT("Run Out Cancel"                ) },	// Fujitsu
		{ 204, SMART_VENDOR_MAXTOR , TEXT("Soft ECC Correction"           ) },	// Maxtor
		{ 204, SMART_VENDOR_FUJITSU, TEXT("Shock Count Write Opern"       ) },	// Fujitsu
		{ 205, SMART_VENDOR_MAXTOR , TEXT("Thermal Aspirates"             ) },	// Maxtor
		{ 205, SMART_VENDOR_FUJITSU, TEXT("Shock Rate Write Opern"        ) },	// Fujitsu
		{ 206, SMART_VENDOR_FUJITSU, TEXT("Flying_Height"                 ) },	// Fujitsu
		{ 207, SMART_VENDOR_MAXTOR , TEXT("Spin High Current"             ) },	// Maxtor
		{ 208, SMART_VENDOR_MAXTOR , TEXT("Spin Buzz"                     ) },	// Maxtor
		{ 209, SMART_VENDOR_MAXTOR , TEXT("Offline Seek Performnce"       ) },	// Maxtor
		{ 220, SMART_VENDOR_GENERAL, TEXT("Disk Shift"                    ) },
		{ 221, SMART_VENDOR_GENERAL, TEXT("G-Sense Error Rate"            ) },
		{ 222, SMART_VENDOR_GENERAL, TEXT("Loaded Hours"                  ) },
		{ 223, SMART_VENDOR_GENERAL, TEXT("Load/Unload Retry Count"       ) },
		{ 224, SMART_VENDOR_GENERAL, TEXT("Load Friction"                 ) },
		{ 225, SMART_VENDOR_GENERAL, TEXT("Load Cycle Count"              ) },
		{ 226, SMART_VENDOR_GENERAL, TEXT("Load-in Time"                  ) },
		{ 227, SMART_VENDOR_GENERAL, TEXT("Torque Amplification Count"    ) },
		{ 228, SMART_VENDOR_GENERAL, TEXT("Power-off Retract Count"       ) },
		{ 230, SMART_VENDOR_GENERAL, TEXT("GMR Head Amplitude"            ) },	// IBM
		{ 231, SMART_VENDOR_GENERAL, TEXT("Temperature Celsius"           ) },
		{ 240, SMART_VENDOR_GENERAL, TEXT("Head Flying Hours"             ) },
		{ 250, SMART_VENDOR_GENERAL, TEXT("Read Error Retry Rate"         ) },
	//	{   0, SMART_VENDOR_GENERAL, TEXT(""                              ) },
	};
	int				i;
	const TCHAR		*text = NULL;

	for ( i = 0; i < sizeof(tbl)/sizeof(*tbl); i++ )
	{
		if ( tbl[i].id == id )
		{
			if ( text == NULL )
				text = tbl[i].text;
			if ( tbl[i].vendor == vendor )
				break;
			break;
		}
	}
	if ( text == NULL )
	{
		static TCHAR	msgbuf[80];

		_stprintf(msgbuf, TEXT("Unknown ID#%3d (%02XH)"), id, id);
		text = msgbuf;
	}
	if ( length > 0 && buffer != NULL )
	{
		if ( length > SMART_ATTRIBUTE_MAX_LENGHTH )
			length = SMART_ATTRIBUTE_MAX_LENGHTH;
		_tcsncpy(buffer, text, length);
		return buffer;
	} else
	{
		return text;
	}
}

DevTest::SmartAttrItem *	DevTest::GetSmartAttrItem(unsigned id)
{
	SmartAttrList::iterator	it;
	SmartAttrItem			*item = NULL;

	for ( it = this->smart.attrList.begin(); it != this->smart.attrList.end(); it++ )
	{
		if ( (*it)->id == id )
		{
			item = *it;
			break;
		}
	}
	return item;
}

const TCHAR * DevTest::GetSmartRawValue(DevCtl::AtaSmartAttribute & attr, unsigned length, TCHAR * buffer)
{
	SmartAttrItem			*item;
	SmartAttrRawType		rawtype = SMART_RAW_NORMAL;
	UINT64					n64;
	TCHAR					tmpbuf[80];

	if ( (item = this->GetSmartAttrItem(attr.id)) != NULL )
		rawtype = item->type;
	n64 = DevTest::ATA_GET48(attr.raw);
	switch ( rawtype )
	{
	case SMART_RAW_DUMPHEX:
		Fmt(tmpbuf, TEXT("%02X %02X %02X %02X %02X %02X "), attr.raw[0], attr.raw[1], attr.raw[2], attr.raw[3], attr.raw[4], attr.raw[5]);
		break;

	case SMART_RAW_DIVWORD:
	{
		unsigned	n1, n2, n3;

		n1 = DevTest::ATA_GET16(&attr.raw[0]);
		n2 = DevTest::ATA_GET16(&attr.raw[2]);
		n3 = DevTest::ATA_GET16(&attr.raw[4]);
		Fmt(tmpbuf, TEXT("%6,u %6,u %6,u "), n1, n2, n3);
		break;
	}

	case SMART_RAW_NORMAL:
	default:
		Fmt(tmpbuf, TEXT("%20,I64u "), n64);
		break;
	}
	_tcsncpy(buffer, tmpbuf, length);

	return buffer;
}

int		DevTest::DisplayInfo(unsigned tblnum, DevTest::DispParam * tbl)
{
	unsigned			i;
	TCHAR				t1[80], t2[80];

	for ( i = 0; i < tblnum; i++ )
	{
		switch ( tbl[i].disptype )
		{
		case DISPTYPE_SHORT_TEXT:
			PutPos(8                                     , tbl[i].name
			      ,COLUMN3 - _tcslen(tbl[i].data.text)-1, tbl[i].data.text
			      ,0);
			break;

		case DISPTYPE_LBA64_CAPACITY:
			PutPos(8         , tbl[i].name
			      ,COLUMN3-21, Fmt(t1, "%20,I64u", *(tbl[i].data.u64))
			      ,COLUMN3   , Fmt(t2, (this->opt.byteUnit == BYTEUNIT_210 ? TEXT("(%I64K)"):TEXT("(%I64k)")), *(tbl[i].data.u64) * I64(512))
			      ,0);
			break;

		case DISPTYPE_UINT32_512SIZE:
			PutPos(8        , tbl[i].name
			      ,COLUMN3-21, Fmt(t1, TEXT("%20,u"), *(tbl[i].data.u32))
			      ,COLUMN3   , Fmt(t2, (this->opt.byteUnit == BYTEUNIT_210 ? TEXT("(%K)"):TEXT("(%k)")), *(tbl[i].data.u32) * 512)
			      ,0);
			break;

		case DISPTYPE_UINT16_BITS:
		{
			int				n;
			unsigned		command;
			TCHAR			bitmsg[20];

			command = *(tbl[i].data.u16);
			for ( n = 15; n >= 0; n-- )
				bitmsg[n] = command & (1<<n) ? TEXT('1') : TEXT('0');
			bitmsg[16] = TEXT('\0');

			PutPos(8, tbl[i].name
			      ,COLUMN3 - _tcslen(bitmsg) - 1, bitmsg
			      ,COLUMN3                      , Fmt(t1, TEXT("%04XH"), command)
			      ,0);
			break;
		}
		case DISPTYPE_UINT16_BITDEF:
		{
			int				n;
			unsigned		command;
			DispParamBitDef	*btbl;
			TCHAR			bitmsg[20];

			command = *(tbl[i].data.u16);
			for ( n = 15; n >= 0; n-- )
				bitmsg[n] = command & (1<<n) ? TEXT('1') : TEXT('0');
			bitmsg[16] = TEXT('\0');

			PutPos(8, tbl[i].name
			      ,COLUMN3 - _tcslen(bitmsg) - 1, bitmsg
			      ,COLUMN3                      , Fmt(t1, TEXT("%04XH"), command)
			      ,0);
			btbl = tbl[i].tbl.bitdef;
			for ( n = 0; n < (int)tbl[i].tblnum; n++ )
			{
				if ( (btbl[n].set && (command & (1<<btbl[n].bit))) || (btbl[n].set == 0 && (command & (1<<btbl[n].bit)) == 0) )
					PutPos(COLUMN3 - _tcslen(btbl[n].text) - 1, btbl[n].text
					      , COLUMN3, Fmt(t1, TEXT("(bit%2u=%u"), btbl[n].bit,  (command>>btbl[n].bit)&1)
					      ,0);
			}
			break;
		}
		}
	}
	return 0;
}
