//////////////////////////////////////////////////////////////////////////
//	"devctl_scsi.cpp"
//------------------------------------------------------------------------
//	$Id: devctl_scsi.cpp 165 2006-03-21 11:56:20Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"devctl.h"
#include	"linux/hdreg.h"

//////////////////////////////////////////////////////////////////////////
//	 ScsiInqData

DevCtl::ScsiInqData::ScsiInqData()
{
	this->pathId     = 0;
	this->targetId   = 0;
	this->lun        = 0;
	this->devClaimed = FALSE;
	this->size       = 0;
	this->data       = NULL;
	this->unknownId  = FALSE;
}

DevCtl::ScsiInqData::~ScsiInqData()
{
	if ( this->data != NULL )
	{
		delete [] this->data;
		this->data = NULL;
	}
}

int		DevCtl::ScsiCreateInqData(DevCtl::ScsiInqData & inqData, int pathId, int targetId, int lun)
{
	SCSIINQUIRY	inq;

	memset(&inq, 0, sizeof(inq));
	switch ( this->htype )
	{
		case HT_SG:
			if ( this->ScsiInquiry(pathId, targetId, lun, -1, sizeof(inq), &inq) < 0 )
				return -1;
			break;
		case HT_BLKDEV:
		case HT_CDROM:
			if ( this->BlkDeviInquiry(inq) < 0 )
				return -1;
			break;
	}
	inqData.pathId     = pathId;
	inqData.targetId   = targetId;
	inqData.lun        = pathId;
	inqData.devClaimed = FALSE;
	inqData.size       = sizeof(inq);
	if ( inqData.data != NULL )
		delete [] inqData.data;
	if ( (inqData.data = new BYTE [sizeof(inq)]) == NULL )
		return -1;
	memcpy(inqData.data, &inq, sizeof(inq));

	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	SCSI

// Check SCSI Condition (sense buffer)
int		DevCtl::ScsiCheckCondition(DevCtl::Sgt & sgt)
{
	if ( sgt.sense[7] < 10 )
		return 0;	// error
	return sgt.sense[12] == 0x23 || sgt.sense[12] == 0x28 ? 1 : 0;
}

int		DevCtl::ScsiInquiry(int pathId, int targetId, int lun, int pageCode, unsigned size, void * buffer)
{
	Sgt		sgt;

	if ( this->fd < 0 || this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_FROM_DEV;
		sgt.hdr.cmd_len         = SCSI_CDB6_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
		sgt.hdr.dxfer_len       = size;
		sgt.hdr.dxferp          = buffer;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0x12; // Inqurity
		sgt.cdb[1]              = (lun << 5) + (pageCode < 0 ? 0 : 1);
		sgt.cdb[2]              = pageCode < 0 ? 0 : pageCode;
	//	sgt.cdb[3]              = 0;
		sgt.cdb[4]              = size;
	//	sgt.cdb[5]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}


int		DevCtl::ScsiTestUnitReady(int pathId, int targetId, int lun)
{
	Sgt		sgt;

	if ( this->fd < 0 || this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_NONE;
		sgt.hdr.cmd_len         = SCSI_CDB6_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
	//	sgt.hdr.dxfer_len       = 0;
	//	sgt.hdr.dxferp          = NULL;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0x00; // TEST UNIT READY
		sgt.cdb[1]              = lun << 5;
	//	sgt.cdb[2]              = 0;
	//	sgt.cdb[3]              = 0;
	//	sgt.cdb[4]              = 0;
	//	sgt.cdb[5]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}

int		DevCtl::ScsiModeSense(int pathId, int targetId, int lun, BOOL blockDesc, unsigned pc, int pageCode, unsigned size, void * buffer)
{
	Sgt		sgt;

	if ( this->fd < 0 || this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_FROM_DEV;
		sgt.hdr.cmd_len         = size < 256 ? SCSI_CDB6_LEN : SCSI_CDB10_LEN;
	//	sgt.hdr.iovec_count     = 0;	// memset takes care of this
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
		sgt.hdr.dxfer_len       = size;
		sgt.hdr.dxferp          = buffer;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout; // 20000 millisecs == 20 seconds
	//	sgt.hdr.flags           = 0;	// take defaults: indirect IO, etc
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		if ( size < 256 )
		{
			sgt.cdb[0]         = 0x1A; // MODE SENSE(6)
			sgt.cdb[1]         = (lun << 5) + (blockDesc ? 0x08:0);
			sgt.cdb[2]         = ((pc&3)<<6) + (pageCode & 0x3F);
		//	sgt.cdb[3]         = 0;
			sgt.cdb[4]         = (BYTE)size;
		//	sgt.cdb[5]         = 0;
		} else
		{
			sgt.cdb[0]         = 0x5A; // MODE SENSE(10)
			sgt.cdb[1]         = (lun << 5) + (blockDesc ? 0x08:0);
			sgt.cdb[2]         = ((pc&3)<<6) + (pageCode & 0x3F);
		//	sgt.cdb[3]         = 0;
		//	sgt.cdb[4]         = 0;
		//	sgt.cdb[5]         = 0;
		//	sgt.cdb[6]         = 0;
			sgt.cdb[7]         = (BYTE)((size >> 8) & 0xFF);
			sgt.cdb[8]         = (BYTE)( size       & 0xFF);
		//	sgt.cdb[9]         = 0;
		}
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}

int		DevCtl::ScsiReadCapacity(int pathId, int targetId, int lun, unsigned size, void * buffer)
{
	Sgt		sgt;

	if ( this->fd < 0 )
		return -1;
	if ( (this->htype == HT_BLKDEV || this->htype == HT_CDROM || this->htype == HT_FILE)
	  && size >= sizeof(SCSICAPACITYDATA) )
	{
		SCSICAPACITYDATA	*capa;

		capa = (SCSICAPACITYDATA *)buffer;
		DevCtl::ScsiSetLBA(capa->logicalBlockAddress, this->blk.blkNum);
		DevCtl::ScsiSetLBA(capa->blockLength        , this->blk.blkLen);
		return 0;
	}
	if ( this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_FROM_DEV;
		sgt.hdr.cmd_len         = SCSI_CDB10_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
		sgt.hdr.dxfer_len       = size;
		sgt.hdr.dxferp          = buffer;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0x25; // Read Capacity
		sgt.cdb[1]              = lun << 5;
	//	sgt.cdb[2]              = 0;
	//	sgt.cdb[3]              = 0;
	//	sgt.cdb[4]              = 0;
	//	sgt.cdb[5]              = 0;
	//	sgt.cdb[6]              = 0;
	//	sgt.cdb[7]              = 0;
	//	sgt.cdb[8]              = 0;	// 0x01 == Num of Sec
	//	sgt.cdb[9]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}

int		DevCtl::ScsiRead(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, void * buffer)
{
	Sgt		sgt;

	if ( this->fd < 0 )
		return -1;
	if ( this->htype == HT_BLKDEV || this->htype == HT_CDROM || this->htype == HT_FILE )
	{
		if ( lseek(this->fd, (INT64)addr * this->blk.blkLen, SEEK_SET) < 0 )
			return -1;
		return read(this->fd, buffer, size);
	}
	if ( this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_FROM_DEV;
		sgt.hdr.cmd_len         = SCSI_CDB10_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
		sgt.hdr.dxfer_len       = size;
		sgt.hdr.dxferp          = buffer;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0x28; // Read
		sgt.cdb[1]              = lun << 5;
		sgt.cdb[2]              = (BYTE)((addr >> 24) & 0xFF);
		sgt.cdb[3]              = (BYTE)((addr >> 16) & 0xFF);
		sgt.cdb[4]              = (BYTE)((addr >>  8) & 0xFF);
		sgt.cdb[5]              = (BYTE)( addr        & 0xFF);
	//	sgt.cdb[6]              = 0;
		sgt.cdb[7]              = (BYTE)((count >> 8) & 0xFF);
		sgt.cdb[8]              = (BYTE)( count       & 0xFF);
	//	sgt.cdb[9]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return (int)size; //sgt.hdr.resid;	// OK
	}
}

	//	sgt.hdr.iovec_count     = 0;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
	//	sgt.cdb[6]              = 0;
	//	sgt.cdb[9]              = 0;

int		DevCtl::ScsiVerify(int pathId, int targetId, int lun, INT64 addr, unsigned count)
{
	Sgt		sgt;

	if ( this->fd < 0 || this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_NONE;
		sgt.hdr.cmd_len         = SCSI_CDB10_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
	//	sgt.hdr.dxfer_len       = 0;
	//	sgt.hdr.dxferp          = NULL;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0x2F; // Verify
		sgt.cdb[1]              = lun << 5;
		sgt.cdb[2]              = (BYTE)((addr >> 24) & 0xFF);
		sgt.cdb[3]              = (BYTE)((addr >> 16) & 0xFF);
		sgt.cdb[4]              = (BYTE)((addr >>  8) & 0xFF);
		sgt.cdb[5]              = (BYTE)( addr        & 0xFF);
	//	sgt.cdb[6]              = 0;
		sgt.cdb[7]              = (BYTE)((count >> 8) & 0xFF);
		sgt.cdb[8]              = (BYTE)( count       & 0xFF);
	//	sgt.cdb[9]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}

	//	sgt.hdr.iovec_count     = 0;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
	//	sgt.cdb[6]              = 0;
	//	sgt.cdb[9]              = 0;

int		DevCtl::ScsiSeek(int pathId, int targetId, int lun, INT64 addr)
{
	Sgt		sgt;

	if ( this->fd < 0 || this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_NONE;
		sgt.hdr.cmd_len         = SCSI_CDB10_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
	//	sgt.hdr.dxfer_len       = 0;
	//	sgt.hdr.dxferp          = NULL;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0x2B; // Seek
		sgt.cdb[1]              = lun << 5;
		sgt.cdb[2]              = (BYTE)((addr >> 24) & 0xFF);
		sgt.cdb[3]              = (BYTE)((addr >> 16) & 0xFF);
		sgt.cdb[4]              = (BYTE)((addr >>  8) & 0xFF);
		sgt.cdb[5]              = (BYTE)( addr        & 0xFF);
	//	sgt.cdb[6]              = 0;
	//	sgt.cdb[7]              = 0;
	//	sgt.cdb[8]              = 0;
	//	sgt.cdb[9]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}

int		DevCtl::ScsiReadToc(int pathId, int targetId, int lun, BOOL msf, unsigned size, void * buffer)
{
	Sgt		sgt;

	if ( this->fd < 0 )
		return -1;
	if ( this->htype == HT_CDROM )
	{
		this->CdromReadToc(msf, size, buffer);
		return 0;
	}

	if ( this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_FROM_DEV;
		sgt.hdr.cmd_len         = SCSI_CDB10_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
		sgt.hdr.dxfer_len       = size;
		sgt.hdr.dxferp          = buffer;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0x43; // Read Toc
		sgt.cdb[1]              = (lun << 5) + (msf ? 0x02 : 0);
	//	sgt.cdb[2]              = 0;
	//	sgt.cdb[3]              = 0;
	//	sgt.cdb[4]              = 0;
	//	sgt.cdb[5]              = 0;
	//	sgt.cdb[6]              = 0;
		sgt.cdb[7]              = (BYTE)((size >> 8) & 0xFF);
		sgt.cdb[8]              = (BYTE)( size       & 0xFF);
	//	sgt.cdb[9]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}

int		DevCtl::ScsiReadToc(int pathId, int targetId, int lun, BOOL msf, DevCtl::SCSITOCDATA & toc)
{
	return this->ScsiReadToc(pathId, targetId, lun, msf, sizeof(toc), &toc);
}

int		DevCtl::ScsiReadDefectData(int pathId, int targetId, int lun, unsigned list, unsigned fmt, unsigned size, void * buffer)
{
	Sgt		sgt;

	if ( this->fd < 0 || this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_FROM_DEV;
		sgt.hdr.cmd_len         = SCSI_CDB10_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
		sgt.hdr.dxfer_len       = size;
		sgt.hdr.dxferp          = (void *)buffer;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0x37; // Read Defect Data
		sgt.cdb[1]              = lun << 5;
		sgt.cdb[2]              = ((list & 3) << 3) + (fmt & 3);	// bit:4==Plsi, bit:3==Glist bit0-2==Format
	//	sgt.cdb[3]              = 0;
	//	sgt.cdb[4]              = 0;
	//	sgt.cdb[5]              = 0;
	//	sgt.cdb[6]              = 0;
		sgt.cdb[7]              = (BYTE)((size >> 8) & 0xFF);
		sgt.cdb[8]              = (BYTE)( size       & 0xFF);
	//	sgt.cdb[9]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}

int		DevCtl::ScsiGetConfiguration(int pathId, int targetId, int lun, unsigned rt, unsigned start, unsigned size, void * buffer)
{
	Sgt		sgt;

	if ( this->fd < 0 || this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_FROM_DEV;
		sgt.hdr.cmd_len         = SCSI_CDB10_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
		sgt.hdr.dxfer_len       = size;
		sgt.hdr.dxferp          = (void *)buffer;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0x46; // Get Configuration
		sgt.cdb[1]              = (lun << 5) + (rt & 3);
		sgt.cdb[2]              = (BYTE)((start >>  8) & 0xFF);
		sgt.cdb[3]              = (BYTE)( start        & 0xFF);
	//	sgt.cdb[4]              = 0;
	//	sgt.cdb[5]              = 0;
		sgt.cdb[6]              = 0;
		sgt.cdb[7]              = (BYTE)((size >> 8) & 0xFF);
		sgt.cdb[8]              = (BYTE)( size       & 0xFF);
	//	sgt.cdb[9]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}

int		DevCtl::ScsiReadDvdStructure(int pathId, int targetId, int lun, unsigned addr, unsigned format, unsigned layer, unsigned size, void * buffer)
{
	Sgt		sgt;

	if ( this->fd < 0 || this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_FROM_DEV;
		sgt.hdr.cmd_len         = SCSI_CDB12_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
		sgt.hdr.dxfer_len       = size;
		sgt.hdr.dxferp          = (void *)buffer;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0xAD; // Read DVD Structure
		sgt.cdb[1]              = lun << 5;
		sgt.cdb[2]             = (BYTE)((addr >> 24) & 0xFF);
		sgt.cdb[3]             = (BYTE)((addr >> 16) & 0xFF);
		sgt.cdb[4]             = (BYTE)((addr >>  8) & 0xFF);
		sgt.cdb[5]             = (BYTE)( addr        & 0xFF);
		sgt.cdb[6]             = (BYTE)layer;
		sgt.cdb[7]             = (BYTE)format;
		sgt.cdb[8]             = (BYTE)((size >> 8) & 0xFF);
		sgt.cdb[9]             = (BYTE)( size       & 0xFF);
	//	sgt.cdb[10]            = 0;
	//	sgt.cdb[11]            = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}

// test/debug
int		DevCtl::ScsiSetCdSpeed(int pathId, int targetId, int lun, unsigned rc, unsigned speed)
{
	Sgt		sgt;

	if ( this->fd < 0 || this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_NONE;
		sgt.hdr.cmd_len         = SCSI_CDB10_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
	//	sgt.hdr.dxfer_len       = 0;
	//	sgt.hdr.dxferp          = NULL;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0xBB; // Set CD Speed
		sgt.cdb[1]              = (lun << 5) & (rc & 3);
		sgt.cdb[2]              = (BYTE)((speed >> 8) & 0xFF);
		sgt.cdb[3]              = (BYTE)( speed       & 0xFF);
	//	sgt.cdb[4]              = 0;
	//	sgt.cdb[5]              = 0;
	//	sgt.cdb[6]              = 0;
	//	sgt.cdb[7]              = 0;
	//	sgt.cdb[8]              = 0;
	//	sgt.cdb[9]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}

int		DevCtl::ScsiStartStopUnit(int pathId, int targetId, int lun, unsigned loadeject_start)
{
	Sgt		sgt;

	if ( this->fd < 0 || this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_NONE;
		sgt.hdr.cmd_len         = SCSI_CDB6_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
	//	sgt.hdr.dxfer_len       = 0;
	//	sgt.hdr.dxferp          = NULL;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0x1B; // START STOP UNIT
		sgt.cdb[1]              = lun << 5;
	//	sgt.cdb[2]              = 0;
	//	sgt.cdb[3]              = 0;
		sgt.cdb[4]              = loadeject_start;
	//	sgt.cdb[5]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}

int		DevCtl::ScsiLoadUnloadMedium(int pathId, int targetId, int lun, unsigned loadunload_start)
{
	Sgt		sgt;

	if ( this->fd < 0 || this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_NONE;
		sgt.hdr.cmd_len         = SCSI_CDB6_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
	//	sgt.hdr.dxfer_len       = 0;
	//	sgt.hdr.dxferp          = NULL;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0xA6; // LOAD UNLOAD MEDIUM
		sgt.cdb[1]              = lun << 5;
	//	sgt.cdb[2]              = 0;
	//	sgt.cdb[3]              = 0;
		sgt.cdb[4]              = loadunload_start;
	//	sgt.cdb[5]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}

int		DevCtl::ScsiPreventAllowMediumRemoval(int pathId, int targetId, int lun, unsigned state)
{
	Sgt		sgt;

	if ( this->fd < 0 || this->htype != HT_SG )
		return -1;
	for (;;)
	{
		memset(&sgt, 0, sizeof(sgt));
		sgt.fd                  = this->fd;
		sgt.hdr.interface_id    = 'S';
		sgt.hdr.dxfer_direction = SG_DXFER_NONE;
		sgt.hdr.cmd_len         = SCSI_CDB6_LEN;
	//	sgt.hdr.iovec_count     = 0;
		sgt.hdr.mx_sb_len       = SCSI_SENSE_LEN;
	//	sgt.hdr.dxfer_len       = 0;
	//	sgt.hdr.dxferp          = NULL;
		sgt.hdr.cmdp            = sgt.cdb;
		sgt.hdr.sbp             = sgt.sense;
		sgt.hdr.timeout         = this->scsi_timeout;
	//	sgt.hdr.flags           = 0;
	//	sgt.hdr.pack_id         = 0;
	//	sgt.hdr.usr_ptr         = NULL;
		sgt.cdb[0]              = 0x1E; // PREVENT ALLOW MEDIUM REMOVAL
		sgt.cdb[1]              = lun << 5;
	//	sgt.cdb[2]              = 0;
	//	sgt.cdb[3]              = 0;
		sgt.cdb[4]              = state;
	//	sgt.cdb[5]              = 0;
		if ( ioctl(sgt.fd, SG_IO, &sgt.hdr) < 0 )
			return -1; // error
		if ( (sgt.hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK )
		{
			if ( sgt.hdr.sb_len_wr > 0 && this->ScsiCheckCondition(sgt) > 0 )
				continue;	// retry
			return -1;
		} else
			return sgt.hdr.resid;	// OK
	}
}


TCHAR *	DevCtl::ScsiGetDeviceType(unsigned devType, TCHAR * buffer)
{
	TCHAR		*s;

	switch ( devType & 0x1F )
	{
		case 0x00: return TEXT("Direct access device" );
		case 0x01: return TEXT("Sequential device"    );
		case 0x02: return TEXT("Printer device"       );
		case 0x03: return TEXT("Processor device"     );
		case 0x04: return TEXT("Write-once device"    );
		case 0x05: return TEXT("CD/DVD-ROM device"    );
		case 0x06: return TEXT("Scaner device"        );
		case 0x07: return TEXT("Optical memory device");
		case 0x08: return TEXT("Medium changer device");
		case 0x09: return TEXT("Communication device" );
		default  :
			if ( buffer == NULL )
				return TEXT("Unkown");
			sprintf(buffer, TEXT("Unkown %02X"), devType & 0x1F);
			return buffer;
	}
	return s;
}


